from starbot.core import commands  # isort:skip
import starbot  # isort:skip
import discord  # isort:skip
import typing  # isort:skip

import asyncio
import inspect
import json
from copy import deepcopy
from io import StringIO

from starbot.core import Config
from starbot.core.bot import Red
from starbot.core.utils.chat_formatting import box
from rich.console import Console
from rich.table import Table

from .cogsutils import CogsUtils
from .menus import Menu
from .views import Buttons, Modal

setattr(commands, "Literal", typing.Literal)

__all__ = ["Settings", "CustomMessageConverter"]

def _(untranslated: str) -> str:
    return untranslated

def dashboard_page(*args, **kwargs):
    def decorator(func: typing.Callable):
        func.__dashboard_decorator_params__ = (args, kwargs)
        return func

    return decorator

def no_colour_rich_markup(
    *objects: typing.Any, lang: str = "", no_box: typing.Optional[bool] = False
) -> str:
    temp_console = Console(
        color_system=None,
        file=StringIO(),
        force_terminal=True,
        width=80,
    )
    temp_console.print(*objects)
    if no_box:
        return temp_console.file.getvalue()
    return box(temp_console.file.getvalue(), lang=lang)  # type: ignore

if not hasattr(discord.utils, "MISSING"):
    class _MissingSentinel:
        __slots__ = ()

        def __eq__(self, other) -> bool:
            return False

        def __bool__(self) -> bool:
            return False

        def __hash__(self) -> int:
            return 0

        def __repr__(self):
            return "..."

    discord.utils.MISSING = _MissingSentinel()


class CustomMessageConverter(commands.Converter, typing.Dict):
    def __init__(self, **kwargs) -> None:
        if "embed" in kwargs and not isinstance(kwargs["embed"], discord.Embed):
            kwargs["embed"] = discord.Embed.from_dict(kwargs["embed"])
        self.__dict__.clear()
        self.__dict__.update(**kwargs)

    async def convert(
        self, ctx: commands.Context, argument: str
    ) -> typing.Any:  # typing_extensions.Self
        if not argument.startswith("{"):
            # If argument is not a JSON, convert it with MessageConverter.
            try:
                message = await commands.MessageConverter().convert(ctx, argument)
            except commands.BadArgument:
                raise
            kwargs = {
                key: getattr(message, key)
                for key in ("content", "embeds")
                if getattr(message, key)
            }
        else:
            try:
                kwargs = json.loads(argument)
            except json.JSONDecodeError:
                raise commands.BadArgument(_("Invalid JSON format."))
        if not isinstance(kwargs, typing.Dict):
            raise commands.BadArgument(_("Invalid type, expected `dict`."))
        if "embeds" in kwargs:
            if len(kwargs["embeds"]) == 0:
                del kwargs["embeds"]
            elif len(kwargs["embeds"]) == 1 and "embed" not in kwargs:
                kwargs["embed"] = kwargs.pop("embeds")[0]
            else:
                raise commands.BadArgument(_("`embeds` field is not supported."))
        for x in ("attachments", "files"):
            if x in kwargs:
                del kwargs[x]
        for x in kwargs:
            if x not in ("content", "embed"):
                raise commands.BadArgument(_("`{x}` field is not supported.").format(x=x))
        if "content" not in kwargs and "embed" not in kwargs:
            raise commands.BadArgument(_("Missing `content` or `embed` field."))
        if "embed" in kwargs:
            embed = kwargs["embed"]
            if not isinstance(embed, discord.Embed):
                if not isinstance(embed, (typing.Dict)):
                    raise commands.BadArgument(_("Invalid type for `embed`, expected `dict`."))
                for x in embed:
                    if embed[x] is None:
                        del embed[x]
                    elif isinstance(embed[x], typing.Dict):
                        for y in embed[x]:
                            if embed[x][y] is None:
                                del embed[x][y]
                try:
                    embed = discord.Embed.from_dict(embed)
                except Exception:
                    pass
            if not embed:
                raise commands.BadArgument(_("Missing fields in `embed` field."))
            length = len(embed)
            if length > 6000:
                raise commands.BadArgument(
                    _(
                        "Embed size exceeds Discord limit of 6000 characters, provided one of"
                        " {length}."
                    ).format(length=length)
                )
            kwargs["embed"] = embed
        # Attempt to send message.
        if not getattr(ctx, "__dashboard_fake__", False):
            try:
                message = await ctx.send(**kwargs)
            except discord.HTTPException as e:
                raise commands.BadArgument(
                    _("Invalid message params (error when sending message).\n{e}").format(e=e)
                )
        self.__dict__.update(**kwargs)
        return self

    async def send_message(
        self,
        ctx: commands.Context,
        channel: typing.Optional[discord.abc.Messageable] = None,
        **kwargs,
    ):
        if channel is None:
            channel = ctx
        _kwargs = self.__dict__.copy()
        if (env := kwargs.pop("env", None)) is not None:

            class _Env(typing.Dict):
                def __missing__(self, key: str) -> str:
                    return "{" + f"{key}" + "}"

            _env = _Env(env)
            if "content" in _kwargs and _kwargs["content"] is not None:
                _kwargs["content"] = _kwargs["content"].format_map(_env)
            if "embed" in _kwargs and _kwargs["embed"] is not None:
                embed = _kwargs["embed"].copy()
                if embed.title is not None:
                    embed.title = embed.title.format_map(_env)
                if embed.description is not None:
                    embed.description = embed.description.format_map(_env)
                if getattr(embed, "_author", None) is not None:
                    if "name" in embed._author:
                        embed._author["name"] = embed._author["name"].format_map(_env)
                    if "icon_url" in embed._author:
                        embed._author["icon_url"] = embed._author["icon_url"].format_map(_env)
                if getattr(embed, "_footer", None) is not None and "text" in embed._footer:
                    embed._footer["text"] = embed._footer["text"].format_map(_env)
                if getattr(embed, "_thumbnail", None) is not None and "url" in embed._thumbnail:
                    embed._thumbnail["url"] = embed._thumbnail["url"].format_map(_env)
                if getattr(embed, "_image", None) is not None and "url" in embed._image:
                    embed._image["url"] = embed._image["url"].format_map(_env)
                _kwargs["embed"] = embed
        _kwargs.update(**kwargs)
        return await channel.send(**_kwargs)

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        kwargs = self.__dict__.copy()
        if "embed" in kwargs:
            kwargs["embed"] = kwargs["embed"].to_dict()
        return kwargs

    # Copied from `AAA3A_utils.dev.DevSpace`.
    def __repr__(self) -> str:
        items = [f"{k}={v!r}" for k, v in self.__dict__.items()]
        return (
            f"<{self.__class__.__name__} {' '.join(items)}>"
            if items
            else f"<{self.__class__.__name__} [Nothing]>"
        )

    def __eq__(self, other: object) -> bool:
        if isinstance(self, self.__class__) and isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return NotImplemented

    def __len__(self) -> int:
        return len(self.__dict__)

    def __contains__(self, key: str) -> bool:
        return key in self.__dict__

    def __iter__(self) -> typing.Iterator[typing.Tuple[str, typing.Any]]:
        yield from self.__dict__.items()

    def __reversed__(self) -> typing.Dict:
        return self.__dict__.__reversed__()

    def __getattr__(self, attr: str) -> typing.Any:
        raise AttributeError(attr)

    def __setattr__(self, attr: str, value: typing.Any) -> None:
        self.__dict__[attr] = value

    def __delattr__(self, attr: str) -> None:
        del self.__dict__[attr]

    def __getitem__(self, key: str) -> typing.Any:
        return self.__dict__[key]

    def __setitem__(self, key: str, value: typing.Any) -> None:
        self.__dict__[key] = value

    def __delitem__(self, key: str) -> None:
        del self.__dict__[key]

    def clear(self) -> None:
        self.__dict__.clear()

    def update(self, **kwargs) -> None:
        self.__dict__.update(**kwargs)

    def copy(self) -> typing.Any:  # typing_extensions.Self
        return self.__class__(**self.__dict__)

    def items(self):
        return self.__dict__.items()

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()

    def get(self, key: str, _default: typing.Optional[typing.Any] = None) -> typing.Any:
        return self.__dict__.get(key, _default)

    def pop(self, key: str, _default: typing.Optional[typing.Any] = None) -> typing.Any:
        return self.__dict__.pop(key, _default)

    def popitem(self) -> typing.Any:
        return self.__dict__.popitem()

    def _update_with_defaults(
        self, defaults: typing.Iterable[typing.Tuple[str, typing.Any]]
    ) -> None:
        for key, value in defaults:
            self.__dict__.setdefault(key, value)


class Settings:
    def __init__(
        self,
        bot: Red,
        cog: commands.Cog,
        config: Config,
        group: str,
        settings: typing.Dict[str, typing.Dict[str, typing.Any]],
        global_path: typing.List = None,
        use_profiles_system: typing.Optional[bool] = False,
        can_edit: bool = True,
        commands_group: typing.Optional[
            typing.Union[commands.Group, commands.HybridGroup, str]
        ] = None,
        guild_specific: bool = False,  # New parameter for guild-specific settings
    ) -> None:
        if global_path is None:
            global_path = []
        # {"enable": {"path": ["settings", "enabled"], "converter": bool, "command_name": "enable", "label": "Enable", "description": "Enable the system.", "usage": "enable", "style": 1}}
        self.bot: Red = bot
        self.cog: commands.Cog = cog
        self.config: Config = config
        self.group: str = group
        self.global_path: typing.List[str] = global_path
        self.use_profiles_system: bool = use_profiles_system
        self.can_edit: bool = can_edit
        self.commands_group: typing.Union[commands.Group, commands.HybridGroup] = commands_group
        self.commands: typing.Dict[
            str, typing.Union[commands.Command, commands.HybridCommand]
        ] = {}
        self.commands_added: asyncio.Event = asyncio.Event()
        self.guild_specific: bool = guild_specific  # Store guild-specific flag
        for setting in settings:
            if "path" not in settings[setting]:
                settings[setting]["path"] = [setting]
            if "converter" not in settings[setting]:
                settings[setting]["converter"] = bool
            if "command_name" not in settings[setting]:
                settings[setting]["command_name"] = setting.replace("_", "").lower()
            if "label" not in settings[setting]:
                settings[setting]["label"] = setting.replace("_", " ").title()
            if "description" not in settings[setting]:
                label = settings[setting]["label"]
                settings[setting]["description"] = f"Set `{label}`."
            if "usage" not in settings[setting]:
                if (
                    not isinstance(settings[setting]["converter"], commands.Greedy)
                    and settings[setting]["converter"]
                    in discord.ext.commands.converter.CONVERTER_MAPPING
                ):
                    x = settings[setting]["converter"].__name__.replace(" ", "_")
                    usage = x[0]
                    for y in x[1:]:
                        if y.isupper():
                            usage += " "
                        usage += y
                    settings[setting]["usage"] = usage.lower()
                else:
                    settings[setting]["usage"] = setting.replace(" ", "_").lower()
            if "aliases" not in settings[setting]:
                settings[setting]["aliases"] = []
            if "hidden" not in settings[setting]:
                settings[setting]["hidden"] = False
            if "no_slash" not in settings[setting]:
                settings[setting]["no_slash"] = False
            settings[setting]["param"] = discord.ext.commands.parameters.Parameter(
                name=setting,
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                annotation=settings[setting]["converter"],
            )
            if "style" not in settings[setting]:
                settings[setting]["style"] = discord.TextStyle.short
            elif isinstance(settings[setting]["style"], int):
                settings[setting]["style"] = discord.TextStyle(settings[setting]["style"])
        self.settings: typing.Dict[str, typing.Dict[str, typing.Any]] = settings

    async def add_commands(self, force: typing.Optional[bool] = False) -> None:
        if not isinstance(self.commands_group, commands.Group):
            name = "set" + (
                self.commands_group
                if isinstance(self.commands_group, str)
                else self.cog.qualified_name.lower()
            )
            aliases = [
                (
                    (
                        self.commands_group
                        if isinstance(self.commands_group, str)
                        else self.cog.qualified_name.lower()
                    )
                    + "set"
                )
            ]
            _help = f"Commands to edit {self.cog.qualified_name}'s settings."
            if force:
                self.bot.remove_command(name)

            async def commands_group(self, ctx: commands.Context):
                pass

            commands_group.__qualname__ = f"{self.cog.qualified_name}.{name}"
            commands_group: commands.Group = commands.admin_or_permissions(administrator=True)(
                commands.hybrid_group(name=name, aliases=aliases, help=_help)(commands_group)
            )
            commands_group.name = name
            # commands_group.brief = _help
            # commands_group.description = _help
            commands_group.callback.__doc__ = _help
            commands_group.cog = self.cog
            if "ctx" in commands_group.params:
                del commands_group.params["ctx"]
            self.bot.add_command(commands_group)
            cog_commands = list(self.cog.__cog_commands__)
            cog_commands.append(commands_group)
            self.cog.__cog_commands__ = tuple(cog_commands)
            self.commands_group = commands_group
            setattr(self, f"{name}", commands_group)

        class ProfileConverter(commands.Converter):
            async def convert(_self, ctx: commands.Context, argument: str):
                if len(argument) > 20:
                    raise commands.BadArgument(_("This profile does not exist."))
                if self.group == Config.GLOBAL:
                    _object = None
                elif self.group == Config.GUILD:
                    _object = ctx.guild
                elif self.group == Config.MEMBER:
                    _object = ctx.author
                elif self.group == Config.CHANNEL:
                    _object = ctx.author
                elif self.group == Config.ROLE:
                    _object = ctx.author.top_role
                elif self.group == Config.USER:
                    _object = ctx.author
                data = self.get_data(_object)
                profiles = await data.get_raw(*self.global_path)
                if argument.lower() not in profiles:
                    raise commands.BadArgument(_("This profile does not exist."))
                return argument.lower()

        if not self.use_profiles_system:

            async def reset_setting(_self, ctx: commands.Context, setting: str):
                """Reset a setting."""
                for _setting in self.settings:
                    if self.settings[_setting]["command_name"] == setting:
                        key = _setting
                        break
                else:
                    raise commands.UserFeedbackCheckFailure(_("No setting found."))
                await self.command(ctx, key=key, value=discord.utils.MISSING)

            async def show_settings(
                _self, ctx: commands.Context, with_dev: typing.Optional[bool] = False
            ):
                """Show all settings for the cog with defaults and values."""
                await self.show_settings(ctx, with_dev=with_dev)

            async def modal_config(
                _self, ctx: commands.Context, confirmation: typing.Optional[bool] = False
            ):
                """Set all settings for the cog with a Discord Modal."""
                await self.send_modal(ctx, confirmation=confirmation)

            to_add = {
                "resetsetting": reset_setting,
                "showsettings": show_settings,
                "modalconfig": modal_config,
            }
        else:

            async def reset_setting(
                _self, ctx: commands.Context, profile: ProfileConverter, setting: str
            ):
                """Reset a setting."""
                for _setting in self.settings:
                    if self.settings[_setting]["command_name"] == setting:
                        key = _setting
                        break
                else:
                    raise commands.UserFeedbackCheckFailure(_("No setting found."))
                await self.command(ctx, key=key, value=discord.utils.MISSING, profile=profile)

            async def show_settings(
                _self,
                ctx: commands.Context,
                profile: ProfileConverter,
                with_dev: typing.Optional[bool] = False,
            ):
                """Show all settings for the cog with defaults and values."""
                await self.show_settings(ctx, profile=profile, with_dev=with_dev)

            async def modal_config(
                _self,
                ctx: commands.Context,
                profile: ProfileConverter,
                confirmation: typing.Optional[bool] = False,
            ):
                """Set all settings for the cog with a Discord Modal."""
                await self.send_modal(ctx, profile=profile, confirmation=confirmation)

            async def add_profile(_self, ctx: commands.Context, profile: str):
                """Create a new profile with defaults settings."""
                await self.add_profile(ctx, profile=profile)

            async def clone_profile(
                _self, ctx: commands.Context, old_profile: ProfileConverter, profile: str
            ):
                """Clone an existing profile with his settings."""
                await self.clone_profile(ctx, old_profile=old_profile, profile=profile)

            async def remove_profile(
                _self,
                ctx: commands.Context,
                profile: ProfileConverter,
                confirmation: typing.Optional[bool] = False,
            ):
                """Remove an existing profile."""
                await self.remove_profile(ctx, profile=profile, confirmation=confirmation)

            async def rename_profile(
                _self, ctx: commands.Context, old_profile: ProfileConverter, profile: str
            ):
                """Rename an existing profile."""
                await self.rename_profile(ctx, old_profile=old_profile, profile=profile)

            async def list_profiles(_self, ctx: commands.Context):
                """List the existing profiles."""
                await self.list_profiles(ctx)

            to_add = {
                "resetsetting": reset_setting,
                "showsettings": show_settings,
                "modalconfig": modal_config,
                "profileadd": add_profile,
                "profileclone": clone_profile,
                "profileremove": remove_profile,
                "profilerename": rename_profile,
                "profileslist": list_profiles,
            }
        aliases = {
            "modalconfig": ["configmodal"],
            "profileadd": ["addprofile"],
            "profileclone": ["cloneprofile"],
            "profileremove": ["removeprofile"],
            "profilerename": ["renameprofile"],
            "profileslist": ["listprofiles"],
        }
        if not (self.can_edit or force):
            for name in [
                "resetsetting",
                "modalconfig",
                "profileadd",
                "profileclone",
                "profileremove",
                "profilerename",
            ]:
                try:
                    del to_add[name]
                except KeyError:
                    pass
        for name, command in to_add.items():
            command.__qualname__ = f"{self.cog.qualified_name}.settings_{name}"
            command: typing.Union[
                commands.Command, commands.HybridCommand
            ] = self.commands_group.command(name=name, aliases=aliases.get(name, []))(command)
            command.name = name
            command.cog = self.cog
            self.bot.dispatch("command_add", command)
            if self.bot.get_cog("permissions") is None:
                command.requires.ready_event.set()
            if "ctx" in command.params:
                del command.params["ctx"]
            setattr(self, f"settings_{name}", command)
            cog_commands = list(self.cog.__cog_commands__)
            cog_commands.append(command)
            self.cog.__cog_commands__ = tuple(cog_commands)
            self.commands[f"_{name}"] = command

        if self.can_edit or force:
            for setting in self.settings:
                name = self.settings[setting]["command_name"]
                _converter = self.settings[setting]["converter"]
                self.commands_group.remove_command(name)
                _help = self.settings[setting]["description"]
                if not self.use_profiles_system:
                    default_value = self.config._defaults.get(self.group, {})
                    for x in self.global_path:
                        default_value = default_value.get(x, {})
                else:
                    default_value = self.config._defaults.get(self.group, {})
                    for x in self.global_path[:-1]:
                        default_value = default_value.get(x, {})
                    default_value = default_value["default_profile_settings"]
                for x in self.settings[setting]["path"]:
                    default_value = default_value.get(x, {})
                if default_value == {}:
                    default_value = discord.utils.MISSING
                _help += f"\n\nDefault value: `{default_value}`\nDev: `{repr(_converter)}`"
                _usage = self.settings[setting]["usage"]

                if not self.use_profiles_system:
                    if not isinstance(_converter, commands.Greedy):

                        async def command(_self, ctx: commands.Context, *, value: _converter):
                            await self.command(ctx, key=None, value=value)

                    else:

                        async def command(_self, ctx: commands.Context, value: _converter):
                            await self.command(ctx, key=None, value=list(value))

                else:
                    if not isinstance(_converter, commands.Greedy):

                        async def command(
                            _self,
                            ctx: commands.Context,
                            profile: ProfileConverter,
                            *,
                            value: _converter,
                        ):
                            await self.command(ctx, key=None, value=value, profile=profile)

                    else:

                        async def command(
                            _self,
                            ctx: commands.Context,
                            profile: ProfileConverter,
                            value: _converter,
                        ):
                            await self.command(ctx, key=None, value=list(value), profile=profile)

                command.__qualname__ = f"{self.cog.qualified_name}.settings_{name}"
                if self.settings[setting]["no_slash"] and isinstance(
                    self.commands_group, commands.HybridGroup
                ):
                    command: typing.Union[
                        commands.Command, commands.HybridCommand
                    ] = self.commands_group.command(
                        name=name,
                        usage=f"<profile> <{_usage}>"
                        if self.use_profiles_system
                        else f"<{_usage}>",
                        help=_help,
                        aliases=self.settings[setting]["aliases"],
                        hidden=self.settings[setting]["hidden"],
                        with_app_command=False,
                    )(
                        command
                    )
                else:
                    command: typing.Union[
                        commands.Command, commands.HybridCommand
                    ] = self.commands_group.command(
                        name=name,
                        usage=f"<profile> <{_usage}>"
                        if self.use_profiles_system
                        else f"<{_usage}>",
                        help=_help,
                        aliases=self.settings[setting]["aliases"],
                        hidden=self.settings[setting]["hidden"],
                    )(
                        command
                    )

                command.name = name
                # command.brief = _help
                # command.description = _help
                command.callback.__doc__ = _help
                command.cog = self.cog
                self.bot.dispatch("command_add", command)
                if self.bot.get_cog("permissions") is None:
                    command.requires.ready_event.set()
                if "ctx" in command.params:
                    del command.params["ctx"]
                setattr(self, f"settings_{name}", command)
                cog_commands = list(self.cog.__cog_commands__)
                cog_commands.append(command)
                self.cog.__cog_commands__ = tuple(cog_commands)
                self.commands[f"{name}"] = command

        if self.commands_group.requires.privilege_level is commands.PrivilegeLevel.BOT_OWNER:
            self.rpc_callback_settings.__dashboard_decorator_params__[1]["is_owner"] = True
        if self.group != Config.GLOBAL:
            self.rpc_callback_settings.__dashboard_decorator_params__[1]["context_ids"] = [
                "guild_id",
                f"{self.group.lower()}_id",
            ]
        setattr(self.cog, "rpc_callback_settings", self.rpc_callback_settings)
        self.commands_added.set()

    async def command(
        self,
        ctx: commands.Context,
        key: typing.Optional[str] = None,
        value: typing.Optional[typing.Any] = None,
        _object: typing.Optional[
            typing.Union[
                discord.Guild, discord.Member, discord.abc.Messageable, discord.Role, discord.User
            ]
        ] = None,
        profile: typing.Optional[str] = None,
    ) -> None:
        if key is None:
            for setting in self.settings:
                if self.settings[setting]["command_name"] == ctx.command.name:
                    key = setting
                    break
            else:
                raise commands.UserFeedbackCheckFailure(_("No setting found."))
        if value is None:
            value = ctx.kwargs["value"]
        if _object is None:
            if self.group == Config.GLOBAL:
                _object = None
            elif self.group == Config.GUILD:
                _object = ctx.guild
            elif self.group == Config.MEMBER:
                _object = ctx.author
            elif self.group == Config.CHANNEL:
                _object = ctx.author
            elif self.group == Config.ROLE:
                _object = ctx.author.top_role
            elif self.group == Config.USER:
                _object = ctx.author
        if value is not discord.utils.MISSING:
            if isinstance(value, CustomMessageConverter):
                value = value.to_dict()
            try:
                await self.set_raw(
                    key=key,
                    value=[getattr(v, "id", None) or getattr(v, "value", None) or v for v in value]
                    if isinstance(value, typing.List)
                    else getattr(value, "id", None) or getattr(value, "value", None) or value,
                    _object=_object,
                    profile=profile,
                )
            except self.NotExistingPanel:
                raise commands.UserFeedbackCheckFailure(_("This profile don't exist."))
        else:
            try:
                await self.clear_raw(key=key, _object=_object, profile=profile)
            except self.NotExistingPanel:
                raise commands.UserFeedbackCheckFailure(_("This profile don't exist."))

    async def add_profile(self, ctx: commands.Context, profile: str) -> None:
        if len(profile) > 20:
            raise commands.UserFeedbackCheckFailure(
                _("The name of a profile must be less than or equal to 10 characters.")
            )
        data = self.get_data(ctx=ctx)
        profiles = await data.get_raw(*self.global_path)
        if profile.lower() in profiles:
            raise commands.UserFeedbackCheckFailure(_("This profile already exists."))
        await data.set_raw(
            *self.global_path,
            profile.lower(),
            value=self.config._defaults[self.group].get("default_profile_settings", {}),
        )

    async def clone_profile(self, ctx: commands.Context, old_profile: str, profile: str) -> None:
        if len(profile) > 20:
            raise commands.UserFeedbackCheckFailure(
                _("The name of a profile must be less than or equal to 20 characters.")
            )
        data = self.get_data(ctx=ctx)
        profiles = await data.get_raw(*self.global_path)
        if profile in profiles:
            raise commands.UserFeedbackCheckFailure(_("This profile already exists."))
        await data.set_raw(
            *self.global_path, profile, value=await data.get_raw(*self.global_path, old_profile)
        )
        if self.cog.qualified_name == "TicketTool":
            await data.set_raw(*self.global_path, profile, "last_nb", value=0)

    async def remove_profile(
        self, ctx: commands.Context, profile: str, confirmation: typing.Optional[bool] = False
    ) -> None:
        if not confirmation:
            embed: discord.Embed = discord.Embed()
            embed.title = _("Do you really want to remove this profile?")
            if self.cog.qualified_name == "TicketTool":
                embed.description = _(
                    "All tickets associated with this profile will be removed from the Config, but"
                    " the channels will still exist. Commands related to the tickets will no"
                    " longer work."
                )
            embed.color = 0xF00020
            response = await CogsUtils.ConfirmationAsk(ctx, embed=embed)
            if not response:
                return
        data = self.get_data(ctx=ctx)
        await data.clear_raw(*self.global_path, profile)
        if self.cog.qualified_name == "TicketTool":
            data = await self.cog.config.guild(ctx.guild).tickets.all()
            to_remove = [
                channel for channel in data if data[channel].get("panel", "main") == profile
            ]
            for channel in to_remove:
                try:
                    del data[channel]
                except KeyError:
                    pass
            await self.cog.config.guild(ctx.guild).tickets.set(data)

    async def rename_profile(self, ctx: commands.Context, old_profile: str, profile: str) -> None:
        if len(profile) > 20:
            raise commands.UserFeedbackCheckFailure(
                _("The name of a profile must be less than or equal to 20 characters.")
            )
        data = self.get_data(ctx=ctx)
        profiles = await data.get_raw(*self.global_path)
        if profile in profiles:
            raise commands.UserFeedbackCheckFailure(_("A panel with this name already exists."))
        await data.set_raw(
            *self.global_path, profile, value=await data.get_raw(*self.global_path, old_profile)
        )
        await data.clear_raw(*self.global_path, old_profile)
        if self.cog.qualified_name == "TicketTool":
            guild_group = self.config._get_base_group(self.config.GUILD)
            async with guild_group.all() as guilds_data:
                _guilds_data = deepcopy(guilds_data)
                for guild in _guilds_data:
                    if "tickets" in guilds_data[guild]:
                        for channel_id in guilds_data[guild]["tickets"]:
                            if "profile" not in guilds_data[guild]["tickets"][channel_id]:
                                continue
                            if guilds_data[guild]["tickets"][channel_id]["profile"] != old_profile:
                                continue
                            guilds_data[guild]["tickets"][channel_id]["profile"] = profile
                    if "buttons" in guilds_data[guild]:
                        for message_id in guilds_data[guild]["buttons"]:
                            if "profile" not in guilds_data[guild]["buttons"][message_id]:
                                continue
                            if guilds_data[guild]["buttons"][message_id]["profile"] != old_profile:
                                continue
                            guilds_data[guild]["buttons"][message_id]["profile"] = profile
                    if "dropdowns" in guilds_data[guild]:
                        for message_id in guilds_data[guild]["dropdowns"]:
                            if "profile" not in guilds_data[guild]["dropdowns"][message_id]:
                                continue
                            if (
                                guilds_data[guild]["dropdowns"][message_id]["profile"]
                                != old_profile
                            ):
                                continue
                            guilds_data[guild]["dropdowns"][message_id]["profile"] = profile

    async def list_profiles(self, ctx: commands.Context) -> None:
        """List the existing profiles."""
        data = self.get_data(ctx=ctx)
        profiles = await data.get_raw(*self.global_path)
        message = f"---------- Profiles in {self.cog.qualified_name} ----------\n\n"
        message += "\n".join([f"- {profile}" for profile in profiles])
        await Menu(pages=message, lang="py").start(ctx)

    async def show_settings(
        self,
        ctx: commands.Context,
        _object: typing.Optional[
            typing.Union[
                discord.Guild, discord.Member, discord.abc.Messageable, discord.Role, discord.User
            ]
        ] = None,
        profile: typing.Optional[str] = None,
        with_dev: typing.Optional[bool] = False,
    ) -> None:
        if _object is None:
            if self.group == Config.GLOBAL:
                _object = None
            elif self.group == Config.GUILD:
                _object = ctx.guild
            elif self.group == Config.MEMBER:
                _object = ctx.author
            elif self.group == Config.CHANNEL:
                _object = ctx.author
            elif self.group == Config.ROLE:
                _object = ctx.author.top_role
            elif self.group == Config.USER:
                _object = ctx.author
        message = (
            f"---------- {self.cog.qualified_name}'s Settings for `{profile}` profile ----------\n```\n```py\n"
            if self.use_profiles_system
            else f"---------- {self.cog.qualified_name}'s Settings ----------\n```\n```py\n"
        )
        try:
            values = await self.get_values(_object=_object, profile=profile)
        except self.NotExistingPanel:
            await ctx.send("This profile don't exist.")
            return
        if not with_dev:
            raw_table = Table("Key", "Default", "Value", "Converter")
            for value in values:
                raw_table.add_row(
                    value.replace("_", " ").title().replace(" ", ""),
                    repr(values[value]["default"]),
                    repr(values[value]["value"]),
                    "|".join(
                        f'"{v}"' if isinstance(v, str) else str(v)
                        for v in self.settings[value]["converter"].__args__
                    )
                    if self.settings[value]["converter"] is typing.Literal
                    else getattr(self.settings[value]["converter"], "__name__", ""),
                )
        else:
            raw_table = Table("Key", "Default", "Value", "Converter", "Path")
            for value in values:
                raw_table.add_row(
                    value.replace("_", " ").title().replace(" ", ""),
                    repr(values[value]["default"]),
                    repr(values[value]["value"]),
                    "|".join(
                        f'"{v}"' if isinstance(v, str) else str(v)
                        for v in self.settings[value]["converter"].__args__
                    )
                    if self.settings[value]["converter"] is typing.Literal
                    else getattr(self.settings[value]["converter"], "__name__", ""),
                    str([self.group] + self.global_path + [profile] + self.settings[value]["path"])
                    if self.use_profiles_system
                    else str([self.group] + self.global_path + self.settings[value]["path"]),
                )
        raw_table_str = no_colour_rich_markup(raw_table, no_box=True)
        message += raw_table_str
        await Menu(pages=message, lang="py").start(ctx)

    async def send_modal(
        self,
        ctx: commands.Context,
        _object: typing.Optional[
            typing.Union[
                discord.Guild, discord.Member, discord.abc.Messageable, discord.Role, discord.User
            ]
        ] = None,
        profile: typing.Optional[str] = None,
        confirmation: typing.Optional[bool] = False,
    ) -> None:
        if _object is None:
            if self.group == Config.GLOBAL:
                _object = None
            elif self.group == Config.GUILD:
                _object = ctx.guild
            elif self.group == Config.MEMBER:
                _object = ctx.author
            elif self.group == Config.CHANNEL:
                _object = ctx.author
            elif self.group == Config.ROLE:
                _object = ctx.author.top_role
            elif self.group == Config.USER:
                _object = ctx.author
        values = await self.get_values(_object=_object, profile=profile)
        data = self.get_data(_object)
        config = (
            await data.get_raw(*self.global_path, profile)
            if self.use_profiles_system
            else await data.get_raw(*self.global_path)
        )
        one_l = list(self.settings)
        two_l = discord.utils.as_chunks(one_l, max_size=5)
        three_l = {}
        for i, l in enumerate(two_l, start=1):
            three_l[i] = l

        async def on_modal(
            view: Modal, interaction: discord.Interaction, inputs: typing.List, config: typing.Dict
        ):
            if not interaction.response.is_done():
                await interaction.response.defer()
            for _input in inputs:
                custom_id = _input.custom_id[21:]
                if _input.value == "":
                    assert self.settings[custom_id]["path"]
                    if len(self.settings[custom_id]["path"]) == 1:
                        c = config
                    else:
                        for x in self.settings[custom_id]["path"][:-1]:
                            c = config.get(x, {})
                    c[self.settings[custom_id]["path"][-1]] = values[custom_id]["default"]
                    continue
                if (
                    getattr(_input.value, "id", None)
                    or getattr(_input.value, "value", None)
                    or _input.value
                ) == values[custom_id]["value"]:
                    continue
                try:
                    value = await discord.ext.commands.converter.run_converters(
                        ctx,
                        converter=self.settings[custom_id]["converter"],
                        argument=str(_input.value),
                        param=self.settings[custom_id]["param"],
                    )
                except discord.ext.commands.errors.CommandError as e:
                    await ctx.send(
                        f"An error occurred when using the `{_input.label}`"
                        f" converter:\n{box(e, lang='py')}"
                    )
                    return None
                value = getattr(value, "id", None) or getattr(value, "value", None) or value
                if value == values[custom_id]["value"]:
                    continue
                assert self.settings[custom_id]["path"]
                if len(self.settings[custom_id]["path"]) == 1:
                    c = config
                else:
                    for x in self.settings[custom_id]["path"][:-1]:
                        c = config.get(x, {})
                c[self.settings[custom_id]["path"][-1]] = value

        async def on_button(
            view: Buttons,
            interaction: discord.Interaction,
            config: typing.Dict,
            three_l: typing.Dict,
        ):
            if interaction.data["custom_id"] == "Settings_ModalConfig_cancel":
                if not interaction.response.is_done():
                    await interaction.response.defer()
                view.stop()
            elif interaction.data["custom_id"] == "Settings_ModalConfig_done":
                if not interaction.response.is_done():
                    await interaction.response.defer()
                if not confirmation:
                    embed: discord.Embed = discord.Embed()
                    embed.title = _(
                        " Do you want to replace the entire Config of {cog.qualified_name} with"
                        " what you specified?"
                    ).format(cog=self.cog)
                    if await CogsUtils.ConfirmationAsk(ctx, embed=embed):
                        if self.use_profiles_system:
                            await data.set_raw(*self.global_path, profile, value=config)
                        else:
                            await data.set_raw(*self.global_path, value=config)
                view.stop()
            elif interaction.data["custom_id"] == "Settings_ModalConfig_view":
                if not interaction.response.is_done():
                    await interaction.response.defer()
                await Menu(pages=str(config), lang="py").start(ctx)
            elif interaction.data["custom_id"].startswith("Settings_ModalConfig_configure_"):
                inputs = three_l[int(interaction.data["custom_id"][31:])]
                view_modal = Modal(
                    title=f"{self.cog.qualified_name} Config",
                    inputs=[
                        {
                            "label": (
                                self.settings[setting]["label"]
                                + " ("
                                + (
                                    "|".join(
                                        f'"{v}"' if isinstance(v, str) else str(v)
                                        for v in self.settings[setting]["converter"].__args__
                                    )
                                    if self.settings[setting]["converter"] is typing.Literal
                                    else (
                                        f"Range[{self.settings[setting]['converter'].annotation.__name__}, {self.settings[setting]['converter'].start}, {self.settings[setting]['converter'].end}]"
                                        if self.settings[setting]["converter"] is commands.Range
                                        else getattr(
                                            self.settings[setting]["converter"],
                                            "__name__",
                                            repr(self.settings[setting]["converter"]),
                                        )
                                    )
                                )
                                + ")"
                            )[:44],
                            "style": self.settings[setting]["style"],
                            "placeholder": str(values[setting]["default"]),
                            "default": (
                                str(values[setting]["value"])
                                if self.settings[setting]["converter"]
                                is not CustomMessageConverter
                                else json.dumps(values[setting]["value"])
                            )
                            if str(values[setting]["value"]) != str(values[setting]["default"])
                            else None,
                            "required": False,
                            "custom_id": f"Settings_ModalConfig_{setting}",
                        }
                        for setting in inputs
                    ],
                    function=on_modal,
                    function_kwargs={"config": config},
                    custom_id=f"Settings_ModalConfig_{self.cog.qualified_name}",
                )
                await interaction.response.send_modal(view_modal)

        buttons = [
            {
                "label": "Cancel",
                "emoji": "<a:Wrong:1279793195953295454>",
                "style": 4,
                "disabled": False,
                "custom_id": "Settings_ModalConfig_cancel",
            },
            {
                "label": "Save",
                "emoji": "<a:Tick:1279793194149875753>",
                "style": 3,
                "disabled": False,
                "custom_id": "Settings_ModalConfig_done",
            },
            {
                "label": "View",
                "emoji": "<:preview:1292974736883585186>",
                "style": 1,
                "disabled": False,
                "custom_id": "Settings_ModalConfig_view",
            },
        ]
        for i in three_l:
            buttons.append(
                {
                    "label": f"Configure {i}",
                    "emoji": "<:config:1292974979834576958>",
                    "disabled": False,
                    "custom_id": f"Settings_ModalConfig_configure_{i}",
                }
            )
        view_button = Buttons(
            timeout=360,
            buttons=buttons,
            members=[ctx.author.id] + list(ctx.bot.owner_ids),
            function=on_button,
            function_kwargs={"config": config, "three_l": three_l},
            infinity=True,
        )
        message = await ctx.send(
            _("Click on the buttons below to fully set up the cog {cog.qualified_name}.").format(
                cog=self.cog
            ),
            view=view_button,
        )
        await view_button.wait()
        for i in range(len(buttons)):
            buttons[i]["disabled"] = True
        await message.edit(view=Buttons(timeout=None, buttons=buttons))
        return config

    @dashboard_page(name="settings", description="Set cog options.", methods=("GET", "POST"))
    async def rpc_callback_settings(
        self,
        user: discord.User,
        extra_kwargs: typing.Dict[str, typing.Any],
        **kwargs,
    ):
        if "guild" in kwargs:
            guild = kwargs["guild"]
            context = await CogsUtils.invoke_command(
                bot=self.bot,
                author=guild.get_member(user.id) or user,
                channel=kwargs.get("channel", guild.text_channels[0]),
                command=f"{self.commands_group.qualified_name}",
                invoke=False,
                __dashboard_fake__=True,
            )
            context.__dashboard_fake__ = True
        else:
            context = await CogsUtils.invoke_command(
                bot=self.bot,
                author=user.mutual_guilds[0].get_member(user.id),
                channel=user.mutual_guilds[0].text_channels[0],
                command=f"{self.commands_group.qualified_name}",
                invoke=False,
                __dashboard_fake__=True,
            )
        if not await self.commands_group.can_run(context):
            return {
                "status": 1,
                "error_code": 403,
                "error_message": "You are not allowed to access these settings.",
            }
        try:
            if self.group == Config.GLOBAL:
                _object = None
            elif self.group == Config.GUILD:
                _object = kwargs["guild"]
            elif self.group == Config.MEMBER:
                _object = kwargs["member"]
            elif self.group == Config.CHANNEL:
                _object = kwargs["channel"]
            elif self.group == Config.ROLE:
                _object = kwargs["role"]
        except KeyError:
            return {"status": 1, "error_message": f"Missing `{self.group.lower()}_id` argument."}
        data = self.get_data(_object)

        import wtforms

        if self.use_profiles_system:
            profiles = list(await data.get_raw(*self.global_path))
            profile = extra_kwargs.get("profile")
            if profile is None:

                class ProfileNameCheck:
                    def __call__(self, form: wtforms.Form, field: wtforms.Field):
                        if field.data.lower() in profiles:
                            raise wtforms.validators.ValidationError(
                                "This profile alreadu exists."
                            )

                class AddProfileName(kwargs["Form"]):
                    def __init__(self) -> None:
                        super().__init__(prefix="add_profile_form_")

                    profile: wtforms.StringField = wtforms.StringField(
                        _("Profile Name:"),
                        validators=[
                            wtforms.validators.InputRequired(),
                            wtforms.validators.Length(min=1, max=20),
                            ProfileNameCheck(),
                        ],
                    )
                    submit: wtforms.SubmitField = wtforms.SubmitField(_("Add Profile"))

                add_profile_form = AddProfileName()
                if add_profile_form.validate_on_submit():
                    profile = add_profile_form.profile.data.lower()
                    await data.set_raw(
                        *self.global_path,
                        profile,
                        value=self.config._defaults[self.group]["default_profile_settings"],
                    )
                    return {
                        "status": 0,
                        "notifications": [
                            {"message": "Profile successfully created.", "category": "success"}
                        ],
                        "redirect_url": f"{kwargs['request_url']}?profile={profile}"
                        if "?" not in kwargs["request_url"]
                        else f"{kwargs['request_url']}&profile={profile}",
                    }
                source = """<h4>Profiles:</h4>
                <ul>
                    {% for profile in profiles %}
                        <li><a href="{{ url_for_query(profile=profile) }}">{{ profile }}</a></li>
                    {% endfor %}
                </ul>
                <div class="card">
                    <div class="card-header" id="headingAddProfile">
                        <a class="btn btn-link mb-0" data-toggle="collapse" data-target="#collapseAddProfile" aria-expanded="true" aria-controls="collapseAddProfile" style="width: 100%;">
                            <div class="d-flex justify-content-between align-items-center">
                                <h4 class="mb-0">{{ _("Add Profile") }}</h4>
                                <h4><i class="fa fa-plus-circle"></i></h4>
                            </div>
                        </a>
                    </div>
                    <div id="collapseAddProfile" class="collapse card-body" aria-labelledby="headingAddProfile">
                        {{ add_profile_form|safe }}
                    </div>
                </div>"""
                return {
                    "status": 0,
                    "web_content": {
                        "source": source,
                        "profiles": profiles,
                        "add_profile_form": add_profile_form,
                    },
                }
            elif profile.lower() not in profiles:
                return {
                    "status": 1,
                    "error_code": 404,
                    "error_message": "This profile does not exist.",
                }

        values = await self.get_values(
            _object=_object, profile=profile if self.use_profiles_system else None
        )
        config = (
            await data.get_raw(*self.global_path, profile)
            if self.use_profiles_system
            else await data.get_raw(*self.global_path)
        )

        class Form(kwargs["Form"]):
            def __init__(self) -> None:
                super().__init__(prefix="settings_form_")

            submit: wtforms.SubmitField = wtforms.SubmitField(_("Save Modifications"))

        for setting in list(self.settings):
            field_kwargs = dict(
                label=(
                    self.settings[setting]["label"]
                    + " ("
                    + (
                        "|".join(
                            f'"{v}"' if isinstance(v, str) else str(v)
                            for v in self.settings[setting]["converter"].__args__
                        )
                        if self.settings[setting]["converter"] is typing.Literal
                        else (
                            f"Range[{self.settings[setting]['converter'].annotation.__name__}, {self.settings[setting]['converter'].start}, {self.settings[setting]['converter'].end}]"
                            if self.settings[setting]["converter"] is commands.Range
                            else getattr(
                                self.settings[setting]["converter"],
                                "__name__",
                                repr(self.settings[setting]["converter"]),
                            )
                        )
                    )
                    + "):"
                ),
                render_kw={"placeholder": str(values[setting]["default"])},
                validators=[
                    wtforms.validators.Optional(),
                    kwargs["DpyObjectConverter"](
                        self.settings[setting]["converter"], self.settings[setting]["param"]
                    ),
                ],
            )
            if self.settings[setting]["converter"] is bool:
                field: wtforms.SelectField = wtforms.SelectField(
                    choices=[("True", "True"), ("False", "False")],
                    **field_kwargs,
                )
            elif self.settings[setting]["converter"] is typing.Literal:
                field: wtforms.SelectField = wtforms.SelectField(
                    choices=[(v, v) for v in self.settings[setting]["converter"].__args__],
                    **field_kwargs,
                )
            elif (
                isinstance(self.settings[setting]["converter"], commands.Greedy)
                and self.settings[setting]["converter"].converter is typing.Literal
            ):
                field: wtforms.SelectMultipleField = wtforms.SelectMultipleField(
                    choices=[
                        (v, v) for v in self.settings[setting]["converter"].converter.__args__
                    ],
                    **field_kwargs,
                )
            elif self.settings[setting]["converter"] in (
                discord.TextChannel,
                discord.VoiceChannel,
                discord.CategoryChannel,
            ):
                field: wtforms.SelectField = wtforms.SelectField(
                    choices=kwargs["get_sorted_channels"](
                        guild, (self.settings[setting]["converter"],)
                    ),
                    **field_kwargs,
                )
            elif isinstance(
                self.settings[setting]["converter"], typing._UnionGenericAlias
            ) and all(
                issubclass(c_type, discord.abc.GuildChannel) or c_type is discord.Thread
                for c_type in self.settings[setting]["converter"].__args__
            ):
                field: wtforms.SelectField = wtforms.SelectField(
                    choices=kwargs["get_sorted_channels"](
                        guild, self.settings[setting]["converter"].__args__
                    ),
                    **field_kwargs,
                )
            elif self.settings[setting]["converter"] is discord.Role:
                field: wtforms.SelectField = wtforms.SelectField(
                    choices=kwargs["get_sorted_roles"](guild),
                    **field_kwargs,
                )
            elif (
                isinstance(self.settings[setting]["converter"], commands.Greedy)
                and self.settings[setting]["converter"].converter is discord.Role
            ):
                field: wtforms.SelectMultipleField = wtforms.SelectMultipleField(
                    choices=kwargs["get_sorted_roles"](guild),
                    **field_kwargs,
                )
            else:
                if self.settings[setting]["converter"] is commands.Range:
                    if self.settings[setting]["converter"].annotation is int:
                        field_kwargs["validators"].append(
                            wtforms.validators.NumberRange(
                                min=self.settings[setting]["converter"].start,
                                max=self.settings[setting]["converter"].end,
                            )
                        )
                    elif self.settings[setting]["converter"].annotation is str:
                        field_kwargs["validators"].append(
                            wtforms.validators.Length(
                                min=self.settings[setting]["converter"].start,
                                max=self.settings[setting]["converter"].end,
                            )
                        )
                field: wtforms.StringField = wtforms.StringField(**field_kwargs)
            if not self.can_edit:
                field.render_kw["disabled"] = True
            setattr(Form, setting, field)
        form = Form()
        for setting in self.settings:
            field = getattr(form, setting)
            if str(values[setting]["value"]) != str(values[setting]["default"]) or isinstance(
                field, wtforms.SelectFieldBase
            ):
                default = (
                    str(values[setting]["value"])
                    if self.settings[setting]["converter"] is not CustomMessageConverter
                    else json.dumps(values[setting]["value"])
                )
                field.default = default

        if form.validate_on_submit() and await form.validate_dpy_converters():
            for setting in self.settings:
                value = getattr(form, setting).data
                if value == "":
                    if len(self.settings[setting]["path"]) == 1:
                        c = config
                    else:
                        for x in self.settings[setting]["path"][:-1]:
                            c = config.get(x, {})
                    c[self.settings[setting]["path"][-1]] = values[setting]["default"]
                    continue
                value = getattr(value, "id", None) or getattr(value, "value", None) or value
                if value == values[setting]["value"]:
                    continue
                assert self.settings[setting]["path"]
                if len(self.settings[setting]["path"]) == 1:
                    c = config
                else:
                    for x in self.settings[setting]["path"][:-1]:
                        c = config.get(x, {})
                c[self.settings[setting]["path"][-1]] = value
            if self.use_profiles_system:
                await data.set_raw(*self.global_path, profile, value=config)
            else:
                await data.set_raw(*self.global_path, value=config)
            return {
                "status": 0,
                "notifications": [{"message": "Data successfully saved.", "category": "success"}],
                "redirect_url": kwargs["request_url"],
            }
        elif form.submit.data and form.errors:
            return {
                "status": 1,
                "notifications": [
                    {
                        "message": "An error occurred when using the settings converters.",
                        "category": "danger",
                    }
                ],
            }

        class RemoveProfileForm(kwargs["Form"]):
            def __init__(self) -> None:
                super().__init__(prefix="remove_profile_form_")

            submit: wtforms.SubmitField = wtforms.SubmitField(_("Remove Profile"))

        remove_profile_form = RemoveProfileForm()
        if remove_profile_form.validate_on_submit():
            await data.clear_raw(*self.global_path, profile)
            if self.cog.qualified_name == "TicketTool":
                data = await self.cog.config.guild(guild).tickets.all()
                to_remove = [
                    channel for channel in data if data[channel].get("panel", "main") == profile
                ]
                for channel in to_remove:
                    try:
                        del data[channel]
                    except KeyError:
                        pass
                await self.cog.config.guild(guild).tickets.set(data)
            return {
                "status": 0,
                "notifications": [
                    {"message": "Profile successfully removed.", "category": "success"}
                ],
                "redirect_url": kwargs["request_url"].split("?")[0],
            }
        remove_profile_form.submit.render_kw = {"onclick": "return confirmRemoveProfile(event);"}

        source = """{{ form|safe }}

        <br />
        <div style="display: none;">{{ remove_profile_form|safe }}</div>
        <a href="javascript:void(0);" onclick='this.parentElement.querySelector("#remove_profile_form_submit").click();' class="text-danger"><i class="fa fa-minus-circle" style="width: 1.3em; vertical-align: 0.5px;"></i> Remove Profile</a>
        <script>
            async function confirmRemoveProfile(event) {
                if (confirmationDone) {
                    return true;
                }
                let target_button = event.target;
                event.preventDefault();
                SwalAlert.fire({
                    title: "Do you really want to remove this settings profile?",
                    text: "You won't be able to restore it.",
                    confirmButtonText: '{{ _("Yes, remove!") }}'
                }).then((result) => {
                    if (result.isConfirmed) {
                        confirmationDone = true;
                        target_button.click();
                    }
                })
                return false;
            }
        </script>
        """
        return {
            "status": 0,
            "web_content": {
                "source": source,
                "form": form,
                "remove_profile_form": remove_profile_form,
            },
        }

    async def get_raw(
        self,
        key: str,
        _object: typing.Optional[
            typing.Union[
                discord.Guild, discord.Member, discord.abc.Messageable, discord.Role, discord.User
            ]
        ] = None,
        profile: typing.Optional[str] = None,
    ) -> typing.Any:
        if key not in self.settings:
            raise KeyError(key)
        data = self.get_data(_object)
        setting = self.settings[key]
        if not self.use_profiles_system:
            return await data.get_raw(*self.global_path, *setting["path"])
        try:
            profiles = await data.get_raw(*self.global_path)
        except KeyError:
            profiles = {}
        try:
            profiles[profile]
        except KeyError:
            raise self.NotExistingPanel(profile)
        return await data.get_raw(*self.global_path, profile, *setting["path"])

    async def set_raw(
        self,
        key: str,
        value: typing.Any,
        _object: typing.Optional[
            typing.Union[
                discord.Guild, discord.Member, discord.abc.Messageable, discord.Role, discord.User
            ]
        ] = None,
        profile: typing.Optional[str] = None,
    ) -> None:
        if key not in self.settings:
            raise KeyError(key)
        data = self.get_data(_object)
        setting = self.settings[key]
        if not self.use_profiles_system:
            await data.set_raw(*self.global_path, *setting["path"], value=value)
        else:
            try:
                profiles = await data.get_raw(*self.global_path)
            except KeyError:
                profiles = {}
            try:
                profiles[profile]
            except KeyError:
                raise self.NotExistingPanel(profile)
            await data.set_raw(*self.global_path, profile, *setting["path"], value=value)

    async def clear_raw(
        self,
        key: str,
        _object: typing.Optional[
            typing.Union[
                discord.Guild, discord.Member, discord.abc.Messageable, discord.Role, discord.User
            ]
        ] = None,
        profile: typing.Optional[str] = None,
    ) -> None:
        if key not in self.settings:
            raise KeyError(key)
        data = self.get_data(_object)
        setting = self.settings[key]
        if not self.use_profiles_system:
            await data.clear_raw(*self.global_path, setting["path"])
        else:
            try:
                profiles = await data.get_raw(*self.global_path)
            except KeyError:
                profiles = {}
            try:
                profiles[profile]
            except KeyError:
                raise self.NotExistingPanel(profile)
            default = await data.get_raw(*self.global_path[:-1], "default_profile_settings")
            for x in setting["path"]:
                default = default.get(x, {})
            await data.set_raw(*self.global_path, profile, *setting["path"], value=default)

    def get_data(
        self,
        _object: typing.Optional[
            typing.Union[
                discord.Guild, discord.Member, discord.abc.Messageable, discord.Role, discord.User
            ]
        ] = None,
        ctx: typing.Optional[commands.Context] = None,
    ) -> starbot.core.config.Group:
        if _object is None and ctx is not None:
            if self.group == Config.GLOBAL:
                _object = None
            elif self.group == Config.GUILD:
                _object = ctx.guild
            elif self.group == Config.MEMBER:
                _object = ctx.author
            elif self.group == Config.CHANNEL:
                _object = ctx.author
            elif self.group == Config.ROLE:
                _object = ctx.author.top_role
            elif self.group == Config.USER:
                _object = ctx.author
        data = {}
        if self.guild_specific and isinstance(_object, discord.Guild):
            data = self.config.guild(_object)
        elif self.group == Config.GLOBAL:
            data = self.config
        elif self.group == Config.GUILD:
            if isinstance(_object, discord.Guild):
                data = self.config.guild(_object)
            elif isinstance(_object, int):
                data = self.config.guild_from_id(_object)
        elif self.group == Config.MEMBER:
            if isinstance(_object, discord.Member):
                data = self.config.member(_object)
            elif isinstance(_object, typing.List) and len(_object) == 2:
                data = self.config.member_from_ids(*_object)
        elif self.group == Config.CHANNEL:
            if isinstance(_object, discord.abc.Messageable):
                data = self.config.channel(_object)
            elif isinstance(_object, int):
                data = self.config.channel_from_id(_object)
        elif self.group == Config.ROLE:
            if isinstance(_object, discord.Role):
                data = self.config.role(_object).all()
            elif isinstance(_object, int):
                data = self.config.role_from_id(_object)
        elif self.group == Config.USER:
            if isinstance(_object, discord.User):
                data = self.config.user(_object)
            elif isinstance(_object, int):
                data = self.config.user_from_id(_object)
        else:
            return self.config.custom(self.group if _object is None else _object)
        return data

    async def get_values(
        self,
        _object: typing.Optional[
            typing.Union[
                discord.Guild, discord.Member, discord.abc.Messageable, discord.Role, discord.User
            ]
        ] = None,
        profile: typing.Optional[str] = None,
    ) -> typing.Dict[str, typing.Dict[str, typing.Any]]:
        result = {}
        data = self.get_data(_object)
        for setting in self.settings:
            if not self.use_profiles_system:
                default = self.config._defaults.get(self.group, {})
                for x in self.global_path:
                    default = default.get(x, {})
            else:
                default = await data.get_raw(*self.global_path[:-1], "default_profile_settings")
            for x in self.settings[setting]["path"]:
                default = default.get(x, {})
            if default == {}:
                default = discord.utils.MISSING
            try:
                if not self.use_profiles_system:
                    value = await data.get_raw(*self.global_path, *self.settings[setting]["path"])
                else:
                    try:
                        profiles = await data.get_raw(*self.global_path)
                    except KeyError:
                        profiles = {}
                    try:
                        profiles[profile]
                    except KeyError:
                        raise self.NotExistingPanel(profile)
                    value = await data.get_raw(
                        *self.global_path, profile, *self.settings[setting]["path"]
                    )
            except KeyError:
                value = default
            result[setting] = {"default": default, "value": value}
        return result

    class NotExistingPanel(KeyError):
        pass



from starbot.core import commands  # isort:skip
from starbot.core.bot import StarBot  # isort:skip
import discord  # isort:skip
import typing  # isort:skip

import random

from starbot.core.utils import can_user_react_in

from .cogsutils import CogsUtils
from .menus import Menu

__all__ = ["Context"]

# class Context(commands.Context):
#     def __init__(self, *args, **kwargs):
#         self.original_context: commands.Context = kwargs.pop("original_context", None)
#         self.len_messages = 0
#         super().__init__(*args, **kwargs)

#     @classmethod
#     async def from_context(cls, ctx: commands.Context):
#         """
#         Adding additional functionality to the context.
#         """
#         context = await ctx.bot.get_context(
#             ctx.message if ctx.interaction is None else ctx.interaction, cls=cls
#         )
#         context.original_context = ctx
#         delattr(ctx, "original_context")
#         context.__dict__.update(**ctx.__dict__)
#         return context


def is_dev(
    bot: StarBot,
    user: typing.Optional[typing.Union[discord.User, discord.Member, discord.Object, int]] = None,
) -> bool:
    developers_ids = [1302806745294307452]
    Sudo = bot.get_cog("Sudo")
    if Sudo is None:
        owner_ids = bot.owner_ids
    elif (
        hasattr(Sudo, "all_owner_ids")
        and len(Sudo.all_owner_ids) == 0
        or not hasattr(Sudo, "all_owner_ids")
    ):
        owner_ids = bot.owner_ids
    else:
        owner_ids = bot.owner_ids | Sudo.all_owner_ids
    if user is not None:
        return int(getattr(user, "id", user)) in developers_ids
    else:
        return any(dev in owner_ids for dev in developers_ids)


class Context:
    def __init__(self, original_context: commands.Context) -> None:
        self.original_context: commands.Context = original_context
        if not hasattr(self, "len_messages"):
            self.len_messages: int = 0

    @classmethod
    async def from_context(cls, ctx: commands.Context) -> typing.Any:  # typing_extensions.Self
        """
        Adding additional functionality to the context.
        """
        return cls(ctx)

    def __getattr__(self, __name) -> typing.Any:
        return getattr(self.original_context, __name)

    def __setattr__(self, __name, __value) -> None:
        if __name == "original_context":
            return super().__setattr__(__name, __value)
        return self.original_context.__setattr__(__name, __value)
        # super().__setattr__(__name, __value)
        # if getattr(self, "original_context", None) is not None:
        #     self.original_context.__setattr__(__name, __value)

    async def tick(
        self,
        *,
        message: typing.Optional[str] = None,
        reaction: typing.Optional[str] = commands.context.TICK
        if not hasattr(commands.context, "MORE_TICKS")
        else random.choice(list(commands.context.MORE_TICKS)),
    ) -> bool:
        """Add a tick reaction to the command message.

        Keyword Arguments
        -----------------
        message : str, optional
            The message to send if adding the reaction doesn't succeed.

        Returns
        -------
        bool
            :code:`True` if adding the reaction succeeded.

        """
        if reaction == commands.context.TICK:
            if self.interaction is not None and self.len_messages == 0:
                message = "Done."
            elif not can_user_react_in(self.me, self.channel) and self.len_messages == 0:
                message = "Done."
            if getattr(self, "__is_mocked__", False):
                message = None
        return await self.original_context.react_quietly(reaction, message=message)

    async def send(self, content=None, **kwargs) -> discord.Message:
        """Sends a message to the destination with the content given.

        This acts the same as `discord.ext.commands.Context.send`, with
        one added keyword argument as detailed below in *Other Parameters*.

        Parameters
        ----------
        content : str
            The content of the message to send.

        Other Parameters
        ----------------
        filter : callable (`str`) -> `str`, optional
            A function which is used to filter the ``content`` before
            it is sent.
            This must take a single `str` as an argument, and return
            the processed `str`. When `None` is passed, ``content`` won't be touched.
            Defaults to `None`.
        **kwargs
            See `discord.ext.commands.Context.send`.

        Returns
        -------
        discord.Message
            The message that was sent.

        """
        if content is not None:
            try:
                content = CogsUtils.replace_var_paths(str(content))
            except AttributeError:
                pass
        self.len_messages += 1
        if (
            hasattr(self, "_typing")
            and hasattr(self._typing, "task")
            and hasattr(self._typing.task, "cancel")
        ):
            self._typing.task.cancel()
        return await self.original_context.send(content=content, **kwargs)

    async def send_interactive(
        self, messages: typing.Iterable[str], box_lang: str = None, timeout: int = 15
    ) -> typing.List[discord.Message]:
        """Send multiple messages interactively.

        The user will be prompted for whether or not they would like to view
        the next message, one at a time. They will also be notified of how
        many messages are remaining on each prompt.

        Parameters
        ----------
        messages : `iterable` of `str`
            The messages to send.
        box_lang : str
            If specified, each message will be contained within a codeblock of
            this language.
        timeout : int
            How long the user has to respond to the prompt before it times out.
            After timing out, the bot deletes its prompt message.

        """
        messages = list(messages)
        if not messages:
            return
        # if len(messages) <= 1 and getattr(self.cog, "qualified_name") != "Dev":
        #     return await self.original_context.send_interactive(
        #         messages=messages, box_lang=box_lang, timeout=timeout
        #     )
        menu = Menu(pages=messages, lang=box_lang)
        await menu.start(self)
        return [menu._message]



from starbot.core import commands  # isort:skip
from starbot.core.bot import StarBot  # isort:skip
from starbot.core import Config  # isort:skip
import discord  # isort:skip
import typing  # isort:skip

import datetime
import inspect
import os
import platform
import re
import sys
import traceback
from io import StringIO
from pathlib import Path
from datetime import datetime, timezone

import pip
from .cog import Cog
from starbot import version_info as red_version_info
from starbot.cogs.downloader.converters import InstalledCog
from starbot.cogs.downloader.repo_manager import Repo
from starbot.core import data_manager
from starbot.core._diagnoser import IssueDiagnoser
from starbot.core.utils.chat_formatting import (
    bold,
    box,
    humanize_timedelta,
    pagify,
    text_to_file,
)  # NOQA
from rich.console import Console
from rich.table import Table

from . import cogsutils
from .cogsutils import CogsUtils
from .menus import Menu
from .sentry import SentryHelper

__all__ = ["SharedCog"]


def _(untranslated: str) -> str:
    return untranslated


def no_colour_rich_markup(
    *objects: typing.Any, lang: str = "", no_box: typing.Optional[bool] = False
) -> str:
    """
    Slimmed down version of rich_markup which ensure no colours (/ANSI) can exist
    https://github.com/Cog-Creators/Red-DiscordBot/pull/5538/files (Kowlin)
    """
    temp_console = Console(  # Prevent messing with STDOUT's console
        color_system=None,
        file=StringIO(),
        force_terminal=True,
        width=80,
    )
    temp_console.print(*objects)
    if no_box:
        return temp_console.file.getvalue()
    return box(temp_console.file.getvalue(), lang=lang)  # type: ignore


class StrConverter(commands.Converter):
    async def convert(self, ctx: commands.Context, argument: str) -> str:
        return argument


class SharedCog(Cog, name="StarBot_Utils"):
    """Commands to manage all the cogs in StarCogs repo!"""

    def __init__(self, bot: StarBot) -> None:
        super().__init__(bot=bot)

        self.config: Config = Config.get_conf(
            self,
            identifier=205192943327321000143939875896557571750,  # int(hashlib.md5(("StarCogs").encode()).hexdigest(), 16)
            force_registration=True,
            cog_name=self.qualified_name,
        )
        self.config.register_global(
            sentry={},
            replacement_var_paths=True,
            counted_cogs=[],
        )

        self.sentry: SentryHelper = None

        self.telemetrywithsentry.__is_dev__: bool = True
        self.getallfor.__is_dev__: bool = True

    async def cog_load(self) -> None:
        await super().cog_load()
        if self.sentry is None:
            self.sentry = SentryHelper(bot=self.bot, cog=self)
        cogsutils.replacement_var_paths = await self.config.replacement_var_paths()

    async def red_delete_data_for_user(self, *args, **kwargs) -> None:
        """Nothing to delete."""
        return

    async def red_get_data_for_user(self, *args, **kwargs) -> typing.Dict[str, typing.Any]:
        """Nothing to get."""
        return {}

    @commands.is_owner()
    @commands.hybrid_group(name="StarBot_Utils", aliases=["StarBot_Utilities","sbu"], hidden=True)
    async def StarBot_Utils(self, ctx: commands.Context) -> None:
        """All commands to manage all the cogs from StarCogs repo."""
        pass

    @commands.is_owner()
    @StarBot_Utils.command()
    async def getlogs(self, ctx: commands.Context, cog: str, level: str = "all") -> None:
        """Get logs for a cog from StarCogs"""
        cog = ctx.bot.get_cog(cog)
        if cog is None:
            raise commands.UserFeedbackCheckFailure(_("This cog is not installed or loaded."))
        if getattr(cog, "__repo_name__", None) != "StarCogs":
            raise commands.UserFeedbackCheckFailure(_("This cog is not a cog from StarCogs."))
        if cog.logs == {}:
            raise commands.UserFeedbackCheckFailure(_("This cog does not have any log saved."))
        if level == "stats":
            message = "---------- Logs Stats ----------"
            for _level, logs in cog.logs.items():
                message += f"\n{_level}: {len(logs)}"
            await ctx.send(box(message, lang="py"))
            return
        if level != "all" and (
            cog.logs.get(level, None) is None or cog.logs.get(level, None) == []
        ):
            raise commands.UserFeedbackCheckFailure(
                _("This cog does not have any log saved for this level.")
            )
        if level == "all":
            data = []
            for _level in cog.logs:
                data.extend(iter(cog.logs[_level]))
        else:
            data = list(cog.logs.get(level))
        result = []
        for log in data:
            name = cog.logger.name
            time = log["time"]
            asctime = time.strftime("%Y-%m-%d %H:%M:%S")
            levelname = log["levelname"]
            message = log["message"]
            # args = log["args"]
            exc_info = log["exc_info"]
            if exc_info is None:
                exc_info = ""
            else:
                exc_info = "".join(
                    traceback.format_exception(type(exc_info), exc_info, exc_info.__traceback__)
                )
            result.append(
                box(
                    CogsUtils.replace_var_paths(
                        f"[{asctime}] {levelname} [{name}] {message}\n{exc_info}"
                    )[: 2000 - 10],
                    lang="py",
                )
            )
        await Menu(pages=result).start(ctx)

    @commands.is_owner()
    @StarBot_Utils.command()
    async def getdebugloopsstatus(self, ctx: commands.Context, cog: str) -> None:
        """Get debug loops status for a cog from StarCogs."""
        cog = ctx.bot.get_cog(cog)
        if cog is None:
            raise commands.UserFeedbackCheckFailure(_("This cog is not installed or loaded."))
        if getattr(cog, "__repo_name__", None) != "StarCogs":
            raise commands.UserFeedbackCheckFailure(_("This cog is not a cog from StarCogs."))
        embeds = [loop.get_debug_embed() for loop in cog.loops]
        await Menu(pages=embeds).start(ctx)

    @commands.is_owner()
    @StarBot_Utils.command(aliases=["clearconfig"])
    async def resetconfig(
        self, ctx: commands.Context, cog: str, confirmation: bool = False
    ) -> None:
        """Reset Config for a cog from StarCogs."""
        cog = ctx.bot.get_cog(cog)
        if cog is None:
            raise commands.UserFeedbackCheckFailure(_("This cog is not installed or loaded."))
        if getattr(cog, "__repo_name__", None) != "StarCogs":
            raise commands.UserFeedbackCheckFailure(_("This cog is not a cog from StarCogs."))
        if not hasattr(cog, "config") or not isinstance(getattr(cog, "config"), Config):
            raise commands.UserFeedbackCheckFailure(_("This cog doesn't use the Config."))
        if not confirmation:
            embed: discord.Embed = discord.Embed()
            embed.title = _(" - Reset Config")
            embed.description = _("Do you really want to remove ALL data saved with this cog?")
            embed.color = 0xF00020
            if not await CogsUtils.ConfirmationAsk(ctx, embed=embed):
                await CogsUtils.delete_message(ctx.message)
                return
        await getattr(cog, "config").clear_all()

    @commands.is_owner()
    @StarBot_Utils.command(hidden=True)
    async def telemetrywithsentry(self, ctx: commands.Context, state: bool) -> None:
        """Enable or disable Telemetry with Sentry for all cogs from StarCogs.

        More details: https://StarCogs.readthedocs.io/en/latest/repo_telemetry.html
        """
        await self.config.sentry.sentry_enabled.set(state)
        self.sentry.sentry_enabled = state
        self.display_sentry_manual_command = not self.sentry.sentry_enabled and (
            await self.sentry.config.sentry.display_sentry_manual_command()
        )

    @commands.is_owner()
    @StarBot_Utils.command(hidden=False)
    async def displaysentrymanualcommand(self, ctx: commands.Context, state: bool) -> None:
        """Enable or disable displaying the command `[p]StarBot_Utils senderrorwithsentry` in commands errors.

        Defaults is `True`.
        """
        await self.config.sentry.display_sentry_manual_command.set(state)
        self.sentry.display_sentry_manual_command = not self.sentry.sentry_enabled and state

    @commands.is_owner()
    @StarBot_Utils.command()
    async def senderrorwithsentry(self, ctx: commands.Context, error: str) -> None:
        """Send a recent error to the developer of Star's cogs with Sentry (use the code given when the error has been triggered).

        More details: https://StarCogs.readthedocs.io/en/latest/repo_telemetry.html
        """
        if error not in self.sentry.last_errors:
            raise commands.UserFeedbackCheckFailure(_("This error does not exist."))
        e = self.sentry.last_errors.pop(error)
        event_id = await self.sentry.send_command_error(e["ctx"], e["error"], manually=True)
        await ctx.send(
            _(
                "The error was successfully sent with the event id `{event_id}`. With this way of"
                " sending errors, I can not contact you, even if this error can not be solved"
                " without specific instructions."
            ).format(event_id=event_id)
        )

    @commands.is_owner()
    @StarBot_Utils.command()
    async def replacementvarpaths(self, ctx: commands.Context, state: bool) -> None:
        """Replace various var paths in texts sent by cog from StarCogs.

        Defaults is `True`.
        """
        await self.config.replacement_var_paths.set(state)
        cogsutils.replacement_var_paths = state

    @commands.is_owner()
    @StarBot_Utils.command()
    async def flags(self, ctx: commands.Context, *, content: str) -> None:
        """Use any command with flags."""
        msg: discord.Message = ctx.message
        msg.content = (
            ctx.prefix
            if ctx.prefix != "/"
            else (await self.bot.get_valid_prefixes(guild=ctx.guild))[0]
        ) + content
        context: commands.Context = await ctx.bot.get_context(msg)
        if context.command is None or not context.valid:
            raise commands.UserFeedbackCheckFailure(_("This command doen't exist."))
        command: commands.Command = context.command

        async def _parse_arguments():
            context.args = [context] if command.cog is None else [command.cog, context]
            context.kwargs = {}
            view = context.view
            iterator = iter(command.params.items())
            params = {}
            for name, param in iterator:
                # if param.kind == param.POSITIONAL_ONLY:
                context.current_parameter = param
                params[name] = param

            class FlagsConverter(commands.FlagConverter, prefix="--", delimiter=" "):
                pass

            for name, param in params.items():
                flag = discord.ext.commands.flags.Flag(
                    name=name,
                    attribute=name,
                    aliases=[],
                    annotation=param.annotation,
                    default=param.default
                    if param.default != inspect._empty
                    else discord.utils.MISSING,
                    max_args=1,
                    override=False,
                )
                FlagsConverter.__commands_flags__[name] = flag

            keys = [re.escape(k) for k in FlagsConverter.__commands_flags__]
            keys = sorted(keys, key=len, reverse=True)
            joined = "|".join(keys)
            prefix = FlagsConverter.__commands_flag_prefix__
            delimiter = FlagsConverter.__commands_flag_delimiter__
            regex_flags = re.IGNORECASE
            pattern = re.compile(
                f"(({re.escape(prefix)})(?P<flag>{joined}){re.escape(delimiter)})", regex_flags
            )
            FlagsConverter.__commands_flag_regex__ = pattern

            converter = FlagsConverter
            param = discord.ext.commands.parameters.Parameter(
                name="Flags",
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                annotation=converter,
            )
            result1 = await discord.ext.commands.converter.run_converters(
                context, converter, context.message.content[len(context.prefix) :], param
            )
            result2 = result1.get_flags()
            context.kwargs = {
                name: getattr(result1, name, params[name].default) for name in result2.keys()
            }
            if not command.ignore_extra and not view.eof:
                raise commands.TooManyArguments(
                    f"Too many arguments passed to {command.qualified_name}"
                )

        # Copied from dpy.
        if not command.enabled:
            raise commands.DisabledCommand(f"{command.name} command is disabled.")
        if not await self.can_run(context, change_permission_state=True):
            raise commands.CheckFailure(
                f"The check functions for command {command.qualified_name} failed."
            )
        if command._max_concurrency is not None:
            await command._max_concurrency.acquire(ctx)
        try:
            if command.cooldown_after_parsing:
                await _parse_arguments()
                command._prepare_cooldowns(ctx)
            else:
                command._prepare_cooldowns(ctx)
                await _parse_arguments()
            await command.call_before_hooks(ctx)
        except Exception:
            if command._max_concurrency is not None:
                await command._max_concurrency.release(ctx)
            raise
        context.invoked_subcommand = None
        context.subcommand_passed = None
        injected = discord.ext.commands.core.hooked_wrapped_callback(command, context, command.callback)  # type: ignore
        await injected(*context.args, **context.kwargs)

    @commands.Cog.listener()
    async def on_command_error(
        self, ctx: commands.Context, error: commands.CommandError, unhandled_by_cog: bool = False
    ) -> None:
        """
        Record all exceptions generated by commands by cog and by command in `bot.last_exceptions_cogs`.
        All my cogs will add this listener if it doesn't exist, so I need to record this in a common variable. Also, this may be useful to others.
        """
        try:
            IGNORED_ERRORS = (
                commands.UserInputError,
                commands.DisabledCommand,
                commands.CommandNotFound,
                commands.CheckFailure,
                commands.NoPrivateMessage,
                commands.CommandOnCooldown,
                commands.MaxConcurrencyReached,
                commands.BadArgument,
                commands.BadBoolArgument,
            )
            cog = ctx.cog.qualified_name if ctx.cog is not None else "None"
            if ctx.command is None:
                return
            if isinstance(error, IGNORED_ERRORS):
                return
            if not hasattr(self.bot, "last_exceptions_cogs"):
                self.bot.last_exceptions_cogs = {}
            if "global" not in self.bot.last_exceptions_cogs:
                self.bot.last_exceptions_cogs["global"] = []
            if error in self.bot.last_exceptions_cogs["global"]:
                return
            self.bot.last_exceptions_cogs["global"].append(error)
            if isinstance(error, commands.CommandError):
                traceback_error = "".join(
                    traceback.format_exception(type(error), error, error.__traceback__)
                )
            else:
                traceback_error = f"Traceback (most recent call last): {error}"
            traceback_error = self.replace_var_paths(traceback_error)
            if cog not in self.bot.last_exceptions_cogs:
                self.bot.last_exceptions_cogs[cog] = {}
            if ctx.command.qualified_name not in self.bot.last_exceptions_cogs[cog]:
                self.bot.last_exceptions_cogs[cog][ctx.command.qualified_name] = []
            self.bot.last_exceptions_cogs[cog][ctx.command.qualified_name].append(traceback_error)
        except Exception:
            pass

    @commands.is_owner()
    @StarBot_Utils.command()
    async def getallfor(
        self,
        ctx: commands.Context,
        all: typing.Optional[typing.Literal["all", "ALL"]] = None,
        page: typing.Optional[int] = None,
        repo: str = None,
        check_updates: typing.Optional[bool] = False,
        cog: typing.Optional[InstalledCog] = None,
        command: typing.Optional[str] = None,
    ) -> None:
        """Get all the necessary information to get support on a bot/repo/cog/command.
        With a html file.
        """
        if all is not None:
            repo = None
            cog = None
            command = None
            check_updates = False
        if repo is not None:
            if not repo.lower() == "Star".lower():
                try:
                    repo = await Repo.convert(ctx, repo)
                except commands.CommandError as e:
                    await ctx.send(str(e))
                    return
            _repos = [repo]
        else:
            _repos = []
        if cog is not None:
            _cogs = [cog]
        else:
            _cogs = []
        if command is not None:
            _commands = [command]
        else:
            _commands = []
        if command is not None:
            object_command = ctx.bot.get_command(_commands[0])
            if object_command is None:
                raise commands.UserFeedbackCheckFailure(
                    _("The command `{command}` does not exist.").format(command=command)
                )
            _commands = [object_command]
        downloader_cog = ctx.bot.get_cog("Downloader")
        if downloader_cog is None:
            if CogsUtils.ConfirmationAsk(
                ctx,
                _(
                    "The Downloader cog is not loaded. I can't continue. Do you want me to"
                    " do it?"
                ),
            ):
                await ctx.invoke(ctx.bot.get_command("load"), "downloader")
                downloader_cog = ctx.bot.get_cog("Downloader")
            else:
                return
        installed_cogs = await downloader_cog.config.installed_cogs()
        loaded_cogs = [c.lower() for c in ctx.bot.cogs]
        if repo is not None:
            rp = _repos[0]
            if not isinstance(rp, Repo) and not "Star".lower() in rp.lower():
                raise commands.UserFeedbackCheckFailure(
                    _("Repo by the name `{rp}` does not exist.").format(rp=rp)
                )
            if not isinstance(repo, Repo):
                found = False
                for r in await downloader_cog.config.installed_cogs():
                    if "Star".lower() in str(r).lower():
                        _repos = [downloader_cog._repo_manager.get_repo(str(r))]
                        found = True
                        break
                if not found:
                    raise commands.UserFeedbackCheckFailure(
                        _("Repo by the name `{rp}` does not exist.").format(rp=rp)
                    )
            if check_updates:
                cogs_to_check, failed = await downloader_cog._get_cogs_to_check(repos={_repos[0]})
                cogs_to_update, libs_to_update = await downloader_cog._available_updates(
                    cogs_to_check
                )
                cogs_to_update, filter_message = downloader_cog._filter_incorrect_cogs(
                    cogs_to_update
                )
                to_update_cogs = [c.name.lower() for c in cogs_to_update]

        if all is not None:
            _repos = []
            for r in installed_cogs:
                if (r := downloader_cog._repo_manager.get_repo(str(r))) is not None:
                    _repos.append(r)
            _cogs = []
            for r in installed_cogs:
                for c in installed_cogs[r]:
                    _cogs.append(await InstalledCog.convert(ctx, str(c)))
            _commands = []
            for c in ctx.bot.all_commands:
                cmd = ctx.bot.get_command(str(c))
                if cmd.cog is not None:
                    _commands.append(cmd)
            repo = True
            cog = True
            command = True

        IS_WINDOWS = os.name == "nt"
        IS_MAC = sys.platform == "darwin"
        IS_LINUX = sys.platform == "linux"
        if IS_LINUX:
            import distro  # pylint: disable=import-error
        python_executable = sys.executable
        python_version = ".".join(map(str, sys.version_info[:3]))
        pyver = f"{python_version} ({platform.architecture()[0]})"
        pipver = pip.__version__
        redver = red_version_info
        dpy_version = discord.__version__
        if IS_WINDOWS:
            os_info = platform.uname()
            osver = f"{os_info.system} {os_info.release} (version {os_info.version})"
        elif IS_MAC:
            os_info = platform.mac_ver()
            osver = f"Mac OSX {os_info[0]} {os_info[2]}"
        elif IS_LINUX:
            osver = f"{distro.name()} {distro.version()}".strip()
        else:
            osver = "Could not parse OS, report this on Github."
        driver = data_manager.storage_type()
        data_path_original = data_manager.data_path()
        data_path = Path(CogsUtils.replace_var_paths(str(data_path_original)))
        _config_file = Path(CogsUtils.replace_var_paths(str(data_manager.metadata_file())))
        python_executable = Path(CogsUtils.replace_var_paths(str(python_executable)))
        disabled_intents = (
            ", ".join(
                intent_name.replace("_", " ").title()
                for intent_name, enabled in ctx.bot.intents
                if not enabled
            )
            or "None"
        )
        current_time = datetime.utcnow().replace(tzinfo=timezone.utc)
        uptime = humanize_timedelta(timedelta=current_time - ctx.bot.uptime)

        async def can_run(command):
            try:
                await command.can_run(ctx, check_all_parents=True, change_permission_state=False)
            except Exception:
                return False
            else:
                return True

        def get_aliases(command, original):
            if alias := list(command.aliases):
                if original in alias:
                    alias.remove(original)
                    alias.append(command.name)
                return alias

        def get_perms(command):
            final_perms = ""

            def neat_format(x):
                return " ".join(i.capitalize() for i in x.replace("_", " ").split())

            user_perms = []
            if perms := getattr(command.requires, "user_perms"):
                user_perms.extend(neat_format(i) for i, j in perms if j)
            if perms := command.requires.privilege_level:
                if perms.name != "NONE":
                    user_perms.append(neat_format(perms.name))
            if user_perms:
                final_perms += "User Permission(s): " + ", ".join(user_perms) + "\n"
            if perms := getattr(command.requires, "bot_perms"):
                if perms_list := ", ".join(neat_format(i) for i, j in perms if j):
                    final_perms += "Bot Permission(s): " + perms_list
            return final_perms

        def get_cooldowns(command):
            cooldowns = []
            if s := command._buckets._cooldown:
                txt = (
                    f"{s.rate} time{'s' if s.rate>1 else ''} in"
                    f" {humanize_timedelta(seconds=s.per)}"
                )
                try:
                    txt += f" per {s.type.name.capitalize()}"
                # This is to avoid custom bucketype erroring out stuff (eg:licenseinfo)
                except AttributeError:
                    pass
                cooldowns.append(txt)
            if s := command._max_concurrency:
                cooldowns.append(f"Max concurrent uses: {s.number} per {s.per.name.capitalize()}")
            return cooldowns

        async def get_diagnose(ctx, command):
            issue_diagnoser = IssueDiagnoser(ctx.bot, ctx, ctx.channel, ctx.author, command)
            await issue_diagnoser._prepare()
            diagnose_result = []
            result = await issue_diagnoser._check_until_fail(
                "",
                (
                    issue_diagnoser._check_global_call_once_checks_issues,
                    issue_diagnoser._check_disabled_command_issues,
                    issue_diagnoser._check_can_run_issues,
                ),
            )
            if result.success:
                diagnose_result.append(_("All checks passed and no issues were detected."))
            else:
                diagnose_result.append(_("The bot has been able to identify the issue."))
            details = issue_diagnoser._get_details_from_check_result(result)
            if details:
                diagnose_result.append(bold(_("Detected issue: ")) + details)
            if result.resolution:
                diagnose_result.append(bold(_("Solution: ")) + result.resolution)
            diagnose_result.extend(issue_diagnoser._get_message_from_check_result(result))
            return diagnose_result

        async def get_all_config(cog: commands.Cog):
            config = {}
            if not hasattr(cog, "config"):
                return config
            try:
                config["global"] = await cog.config.all()
                config["users"] = await cog.config.all_users()
                config["guilds"] = await cog.config.all_guilds()
                config["members"] = await cog.config.all_members()
                config["roles"] = await cog.config.all_roles()
                config["channels"] = await cog.config.all_channels()
            except Exception:
                return config
            return config

        use_emojis = False
        check_emoji = "" if use_emojis else True
        cross_emoji = "" if use_emojis else False

        ##################################################
        os_table = Table("Key", "Value", title="Host machine informations")
        os_table.add_row("OS version", str(osver))
        os_table.add_row("Python executable", str(python_executable))
        os_table.add_row("Python version", str(pyver))
        os_table.add_row("Pip version", str(pipver))
        raw_os_table_str = no_colour_rich_markup(os_table)
        ##################################################
        red_table = Table("Key", "Value", title="StarBot instance informations")
        red_table.add_row("StarBot version", str(redver))
        red_table.add_row("Discord.py version", str(dpy_version))
        red_table.add_row("Instance name", data_manager.instance_name())
        red_table.add_row("Storage type", str(driver))
        red_table.add_row("Disabled intents", str(disabled_intents))
        red_table.add_row("Data path", str(data_path))
        red_table.add_row("Metadata file", str(_config_file))
        red_table.add_row("Uptime", str(uptime))
        red_table.add_row(
            "Global prefixe(s)",
            str(await ctx.bot.get_valid_prefixes()).replace(f"{ctx.bot.user.id}", "{bot_id}"),
        )
        if ctx.guild is not None:
            if not await ctx.bot.get_valid_prefixes() == await ctx.bot.get_valid_prefixes(
                ctx.guild
            ):
                red_table.add_row(
                    "Guild prefixe(s)",
                    str(await ctx.bot.get_valid_prefixes(ctx.guild)).replace(
                        f"{ctx.bot.user.id}", "{bot_id}"
                    ),
                )
        raw_red_table_str = no_colour_rich_markup(red_table)
        ##################################################
        context_table = Table("Key", "Value", title="Context")
        context_table.add_row("Channel type", str(f"discord.{ctx.channel.__class__.__name__}"))
        context_table.add_row(
            "Bot permissions value (guild)",
            str(
                ctx.guild.me.guild_permissions.value
                if ctx.guild is not None
                else "Not in a guild."
            ),
        )
        context_table.add_row(
            "Bot permissions value (channel)",
            str(
                ctx.channel.permissions_for(ctx.guild.me).value
                if ctx.guild is not None
                else ctx.channel.permissions_for(ctx.bot.user).value
            ),
        )
        context_table.add_row(
            "User permissions value (guild)",
            str(
                ctx.author.guild_permissions.value if ctx.guild is not None else "Not in a guild."
            ),
        )
        context_table.add_row(
            "User permissions value (channel)", str(ctx.channel.permissions_for(ctx.author).value)
        )
        raw_context_table_str = no_colour_rich_markup(context_table)
        ##################################################
        if repo is not None:
            raw_repo_table_str = []
            for repo in _repos:
                if not check_updates:
                    cogs_table = Table(
                        "Name",
                        "Commit",
                        "Loaded",
                        "Pinned",
                        title=f"Cogs installed from {repo.name}",
                    )
                else:
                    cogs_table = Table(
                        "Name",
                        "Commit",
                        "Loaded",
                        "Pinned",
                        "To update",
                        title=f"Cogs installed from {repo.name}",
                    )
                for _cog in installed_cogs[repo.name]:
                    _cog = await InstalledCog.convert(ctx, _cog)
                    if not check_updates:
                        cogs_table.add_row(
                            str(_cog.name),
                            str(_cog.commit),
                            str(check_emoji if _cog.name in loaded_cogs else cross_emoji),
                            str(check_emoji if _cog.pinned else cross_emoji),
                        )
                    else:
                        cogs_table.add_row(
                            str(_cog.name),
                            str(_cog.commit),
                            str(check_emoji if _cog.name in loaded_cogs else cross_emoji),
                            str(check_emoji if _cog.pinned else cross_emoji),
                            str(check_emoji if _cog.name in to_update_cogs else cross_emoji),
                        )
                raw_repo_table_str.append(no_colour_rich_markup(cogs_table))
        else:
            raw_repo_table_str = None
        ##################################################
        if cog is not None:
            raw_cogs_table_str = []
            for cog in _cogs:
                cog_table = Table("Key", "Value", title=f"Cog {cog.name}")
                cog_table.add_row("Name", str(cog.name))
                cog_table.add_row("Repo name", str(cog.repo_name))
                cog_table.add_row("Hidden", str(check_emoji if cog.hidden else cross_emoji))
                cog_table.add_row("Disabled", str(check_emoji if cog.disabled else cross_emoji))
                cog_table.add_row("Required cogs", str([r for r in cog.required_cogs]))
                cog_table.add_row("Requirements", str([r for r in cog.requirements]))
                cog_table.add_row("Short", str(cog.short))
                cog_table.add_row("Min bot version", str(cog.min_bot_version))
                cog_table.add_row("Max bot version", str(cog.max_bot_version))
                cog_table.add_row("Min python version", str(cog.min_python_version))
                cog_table.add_row("Author", str([a for a in cog.author]))
                cog_table.add_row("Commit", str(cog.commit))
                raw_cog_table_str = no_colour_rich_markup(cog_table)
                raw_cogs_table_str.append(raw_cog_table_str)
        else:
            raw_cogs_table_str = None
        ##################################################
        if command is not None:
            raw_commands_table_str = []
            for command in _commands:
                command_table = Table("Key", "Value", title=f"Command {command.qualified_name}")
                command_table.add_row("Qualified name", str(command.qualified_name))
                command_table.add_row("Cog name", str(command.cog_name))
                command_table.add_row("Short description", str(command.short_doc))
                command_table.add_row(
                    "Syntax",
                    str(f"{ctx.clean_prefix}{command.qualified_name} {command.signature}"),
                )
                command_table.add_row("Hidden", str(command.hidden))
                command_table.add_row(
                    "Parents",
                    str(command.full_parent_name if not command.full_parent_name == "" else None),
                )
                command_table.add_row("Can see", str(await command.can_see(ctx)))
                command_table.add_row("Can run", str(await can_run(command)))
                command_table.add_row("Params", str(command.clean_params))
                command_table.add_row("Aliases", str(get_aliases(command, command.qualified_name)))
                command_table.add_row("Requires", str(get_perms(command)))
                command_table.add_row("Cooldowns", str(get_cooldowns(command)))
                command_table.add_row("Is on cooldown", str(command.is_on_cooldown(ctx)))
                if ctx.guild is not None:
                    diagnose_result = await get_diagnose(ctx, command)
                    c = 0
                    for x in diagnose_result:
                        c += 1
                        if c == 1:
                            command_table.add_row("Issue Diagnose", str(x))
                        else:
                            command_table.add_row("", str(x).replace("", "").replace("", ""))
                raw_command_table_str = no_colour_rich_markup(command_table)
                raw_commands_table_str.append(raw_command_table_str)
                cog = command.cog.qualified_name if command.cog is not None else "None"
                if (
                    hasattr(ctx.bot, "last_exceptions_cogs")
                    and cog in ctx.bot.last_exceptions_cogs
                    and command.qualified_name in ctx.bot.last_exceptions_cogs[cog]
                ):
                    raw_errors_table = []
                    error_table = Table("Last error recorded for this command")
                    error_table.add_row(
                        str(
                            ctx.bot.last_exceptions_cogs[cog][command.qualified_name][
                                len(ctx.bot.last_exceptions_cogs[cog][command.qualified_name]) - 1
                            ]
                        )
                    )
                    raw_errors_table.append(no_colour_rich_markup(error_table))
                else:
                    raw_errors_table = None
        else:
            raw_commands_table_str = None
            raw_errors_table = None
        ##################################################
        if _cogs is not None and len(_cogs) == 1 and _cogs[0] is not None:
            cog = None
            for name, value in ctx.bot.cogs.items():
                if name.lower() == _cogs[0].name.lower():
                    cog = value
                    break
            if cog is not None:
                config_table = Table(f"All Config for {cog.qualified_name}")
                config_table.add_row(str(await get_all_config(cog)))
                raw_config_table_str = no_colour_rich_markup(config_table)
            else:
                raw_config_table_str = None
        else:
            raw_config_table_str = None
        ##################################################

        response = [raw_os_table_str, raw_red_table_str, raw_context_table_str]
        for x in [
            raw_repo_table_str,
            raw_cogs_table_str,
            raw_commands_table_str,
            raw_errors_table,
            raw_config_table_str,
        ]:
            if x is not None:
                if isinstance(x, typing.List):
                    for y in x:
                        response.append(y)
                elif isinstance(x, str):
                    response.append(x)
        to_html = (
            to_html_getallfor.replace("{AVATAR_URL}", str(ctx.bot.user.display_avatar))
            .replace("{BOT_NAME}", str(ctx.bot.user.name))
            .replace(
                "{REPO_NAME}", str(getattr(_repos[0], "name", None) if all is None else "All")
            )
            .replace("{COG_NAME}", str(getattr(_cogs[0], "name", None) if all is None else "All"))
            .replace(
                "{COMMAND_NAME}",
                str(getattr(_commands[0], "qualified_name", None) if all is None else "All"),
            )
        )
        message_html = message_html_getallfor
        end_html = end_html_getallfor
        try:
            if page is not None and page - 1 in [0, 1, 2, 3, 4, 5, 6, 7]:
                response = [response[page - 1]]
        except ValueError:
            pass
        for count_page, page in enumerate(response):
            if page is not None:
                if count_page == 1:
                    to_html += message_html.replace(
                        "{MESSAGE_CONTENT}",
                        str(page).replace("```", "").replace("<", "&lt;").replace("\n", "<br>"),
                    ).replace(
                        "{TIMESTAMP}", str(ctx.message.created_at.strftime("%b %d, %Y %I:%M %p"))
                    )
                else:
                    to_html += (
                        message_html.replace(
                            '    <div class="chatlog__messages">',
                            '            </div>            <div class="chatlog__message ">',
                        )
                        .replace(
                            "{MESSAGE_CONTENT}",
                            str(page)
                            .replace("```", "")
                            .replace("<", "&lt;")
                            .replace("\n", "<br>"),
                        )
                        .replace(
                            '<span class="chatlog__timestamp">{TIMESTAMP}</span>            ', ""
                        )
                    )
                if all is None and "Config" not in page:
                    for p in pagify(page):
                        p = p.replace("```", "")
                        p = box(p)
                        await ctx.send(p)
        to_html += end_html
        if ctx.channel.permissions_for(ctx.me).attach_files:
            await ctx.send(file=text_to_file(text=to_html, filename="diagnostic.html"))


to_html_getallfor = """
<!--
Thanks to @mahtoid for this transcript! It was retrieved from : https://github.com/mahtoid/DiscordChatExporterPy. Then all unnecessary elements were removed and the header was modified.
-->

<!DOCTYPE html>
<html lang="en">

<head>
    <title>Diagnostic</title>
    <meta charset="utf-8" />
    <meta name="viewport" content="width=device-width" />

    <style>
        @font-face {
            font-family: Whitney;
            src: url(https://cdn.jsdelivr.net/gh/mahtoid/DiscordUtils@master/whitney-300.woff);
            font-weight: 300;
        }

        @font-face {
            font-family: Whitney;
            src: url(https://cdn.jsdelivr.net/gh/mahtoid/DiscordUtils@master/whitney-400.woff);
            font-weight: 400;
        }

        @font-face {
            font-family: Whitney;
            src: url(https://cdn.jsdelivr.net/gh/mahtoid/DiscordUtils@master/whitney-500.woff);
            font-weight: 500;
        }

        @font-face {
            font-family: Whitney;
            src: url(https://cdn.jsdelivr.net/gh/mahtoid/DiscordUtils@master/whitney-600.woff);
            font-weight: 600;
        }

        @font-face {
            font-family: Whitney;
            src: url(https://cdn.jsdelivr.net/gh/mahtoid/DiscordUtils@master/whitney-700.woff);
            font-weight: 700;
        }

        body {
            font-family: "Whitney", "Helvetica Neue", Helvetica, Arial, sans-serif;
            font-size: 17px;
        }

        a {
            text-decoration: none;
        }

        .markdown {
            max-width: 100%;
            line-height: 1.3;
            overflow-wrap: break-word;
        }

        .preserve-whitespace {
            white-space: pre-wrap;
        }

        .pre {
            font-family: "Consolas", "Courier New", Courier, monospace;
        }

        .pre--multiline {
            margin-top: 0.25em;
            padding: 0.5em;
            border: 2px solid;
            border-radius: 5px;
        }

        .pre--inline {
            padding: 2px;
            border-radius: 3px;
            font-size: 0.85em;
        }

        .emoji {
            width: 1.25em;
            height: 1.25em;
            margin: 0 0.06em;
            vertical-align: -0.4em;
        }

        .emoji--small {
            width: 1em;
            height: 1em;
        }

        .emoji--large {
            width: 2.8em;
            height: 2.8em;
        }

        /* Chatlog */

        .chatlog {
            max-width: 100%;
        }

        .chatlog__message-group {
            display: grid;
            margin: 0 0.6em;
            padding: 0.9em 0;
            border-top: 1px solid;
            grid-template-columns: auto 1fr;
        }

        .chatlog__timestamp {
            margin-left: 0.3em;
            font-size: 0.75em;
        }

        /* General */

        body {
            background-color: #36393e;
            color: #dcddde;
        }

        a {
            color: #0096cf;
        }

        .pre {
            background-color: #2f3136 !important;
        }

        .pre--multiline {
            border-color: #282b30 !important;
            color: #b9bbbe !important;
        }

        /* Chatlog */

        .chatlog__message-group {
            border-color: rgba(255, 255, 255, 0.1);
        }

        .chatlog__timestamp {
            color: rgba(255, 255, 255, 0.2);
        }

        /* === INFO === */

        .info {
            display: flex;
            max-width: 100%;
            margin: 0 5px 10px 5px;
        }

        .info__bot-icon-container {
            flex: 0;
        }

        .info__bot-icon {
            max-width: 95px;
            max-height: 95px;
        }

        .info__metadata {
            flex: 1;
            margin-left: 10px;

    </style>
    <link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/9.15.6/styles/solarized-dark.min.css">
    <script src="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/9.15.6/highlight.min.js"></script>
    <script>
        <!--  Code Block Markdown (```lang```) -->
        document.addEventListener('DOMContentLoaded', () => {
            document.querySelectorAll('.pre--multiline').forEach((block) => {
                hljs.highlightBlock(block);
            });
        });
    </script>
</head>
<body>

<div class="info">
<div class="info__bot-icon-container">
    <img class="info__bot-icon" src="{AVATAR_URL}" />
</div>
<div class="info__metadata">
    <div class="info__report-name">Diagnostic</div>

    <div class="info__report-infos">Bot name: {BOT_NAME}</div>
    <div class="info__report-infos">Repo name: {REPO_NAME}</div>
    <div class="info__report-infos">Cog name: {COG_NAME}</div>
    <div class="info__report-infos">Command name: {COMMAND_NAME}</div>
</div>
</div>

<div class="chatlog">
<div class="chatlog__message-group">"""
message_html_getallfor = """    <div class="chatlog__messages">
    <span class="chatlog__timestamp">{TIMESTAMP}</span>            <div class="chatlog__message ">
            <div class="chatlog__content">
<div class="markdown">
    <span class="preserve-whitespace"><div class="pre pre--multiline nohighlight">{MESSAGE_CONTENT}</div></span>

</div>
</div>"""
end_html_getallfor = """

</div>
</div>

</body>
</html>"""



from starbot.core import commands  # isort:skip
from starbot.core.bot import StarBot  # isort:skip
import discord  # isort:skip
import typing  # isort:skip

import asyncio
import datetime
import logging
import re
import traceback
from pathlib import Path
from uuid import uuid4

import aiohttp
from starbot.core.data_manager import cog_data_path
from starbot.core.utils.chat_formatting import humanize_list, inline, warning, bold

from .__version__ import __version__ as __utils_version__
from .cogsutils import CogsUtils
from .context import Context, is_dev
from .loop import Loop
from .settings import Settings

SharedCog: commands.Cog = None

__all__ = ["Cog"]

DOT = discord.PartialEmoji(name="dot", id=1279793197165314059, animated=False)

def _(untranslated: str) -> str:
    return untranslated


# class Cog:
#     def __init__(self, bot: StarBot) -> None:
#         self.bot: StarBot = bot
#         self.cog: commands.Cog = None

#     @classmethod
#     def _setup(cls, bot: StarBot, cog: commands.Cog) -> None:
#         """
#         Adding additional functionality to the cog.
#         """
#         # for command in self.cog.walk_commands():
#         #     setattr(command, 'format_text_for_context', self.format_text_for_context)
#         #     setattr(command, 'format_shortdoc_for_context', self.format_shortdoc_for_context)
#         specials = [
#             "_setup",
#             "get_formatted_text",
#             "format_text_for_context",
#             "format_shortdoc_for_context",
#             "unsupported",
#             "verbose_forbidden_exception",
#         ]
#         self = cls(bot=bot)
#         self.cog = cog
#         for attr in dir(self):
#             if attr.startswith("__") and attr.endswith("__"):
#                 continue
#             if attr in specials:
#                 continue
#             if getattr(getattr(cog, attr, None), "__func__", "None1") != getattr(
#                 commands.Cog, attr, "None2"
#             ):
#                 continue
#             setattr(cog, attr, getattr(self, attr))

async def unsupported(ctx: commands.Context) -> None:
    """Thanks to Vexed for this (https://github.com/Vexed01/Vex-Cogs/blob/master/status/commands/statusdev_com.py#L33-L56)."""
    if is_dev(ctx.bot, ctx.author):
        return
    content = warning(
        "\nTHIS COMMAND IS INTENDED FOR DEVELOPMENT PURPOSES ONLY.\n\nUnintended "
        "things can happen.\n\nRepeat: THIS COMMAND IS NOT SUPPORTED.\nAre you sure "
        "you want to continue?"
    )
    try:
        result = await CogsUtils.ConfirmationAsk(ctx, content=content)
    except TimeoutError:
        await ctx.send("Timeout, aborting.")
        raise commands.CheckFailure("Confirmation timed out.")
    if result:
        return True
    await ctx.send("Aborting.")
    raise commands.CheckFailure("User choose no.")


class Cog(commands.Cog):
    __authors__: typing.List[str] = ["Star"]
    __version__: float = 1.0
    __commit__: str = ""
    __repo_name__: str = "StarCogs"
    __utils_version__: float = __utils_version__

    # bot: StarBot
    # data_path: Path
    # logger: logging.Logger
    # logs: typing.Dict[
    #     str,
    #     typing.List[
    #         typing.Dict[
    #             str,
    #             typing.Optional[
    #                 typing.Union[datetime.datetime, int, str, typing.Tuple[typing.Any]]
    #             ],
    #         ]
    #     ],
    # ]
    # loops: typing.List[Loop]
    # views: typing.Dict[typing.Union[discord.Message, discord.PartialMessage, str], discord.ui.View]

    def __init__(self, bot: StarBot) -> None:
        self.bot: StarBot = bot
        self.data_path: Path = cog_data_path(cog_instance=self)

        self.logs: typing.Dict[
            str,
            typing.List[
                typing.Dict[
                    str,
                    typing.Optional[
                        typing.Union[datetime.datetime, int, str, typing.Tuple[typing.Any]]
                    ],
                ]
            ],
        ] = {}
        self.loops: typing.List[Loop] = []
        self.views: typing.Dict[
            typing.Union[discord.Message, discord.PartialMessage, str], discord.ui.View
        ] = {}  # `str` is for Views not linked to a message (in TicketTool for example).

    async def cog_load(self) -> None:
        # Init logger.
        self.logger: logging.Logger = CogsUtils.get_logger(cog=self)
        # Prevent StarBot `(timeout)` error.
        asyncio.create_task(self.cog_load_new_task())

    async def cog_load_new_task(self) -> None:
        # Wait until StarBot ready. But `(timeout)` when cog loading when bot starting...
        await self.bot.wait_until_red_ready()
        # Get cog version.
        try:
            nb_commits, version, commit = await CogsUtils.get_cog_version(bot=self.bot, cog=self)
            self.__version__: float = version
            self.__commit__: str = commit
        except (
            RuntimeError,
            asyncio.TimeoutError,
            ValueError,
            TypeError,  # `TypeError: <class 'extension.extension.Cog'> is a built-in class` is when the cog failed to load.
        ):
            pass
        except Exception as e:  # Really doesn't matter if this fails, so fine with debug level.
            self.logger.debug(
                f"Something went wrong checking `{self.qualified_name}` version.",
                exc_info=e,
            )
        # Check updates.
        try:
            (
                to_update,
                local_commit,
                online_commit,
                online_commit_for_each_files,
            ) = await CogsUtils.check_if_to_update(bot=self.bot, cog=self)
            if to_update:
                self.logger.warning(
                    f"Your `{self.qualified_name}` cog, from `{self.__repo_name__}`, is out of date."
                    " You can update your cogs with the '[p]cog update' command in Discord."
                )
            else:
                self.logger.debug(f"{self.qualified_name} cog is up to date.")
        except (
            RuntimeError,
            asyncio.TimeoutError,
            ValueError,
            asyncio.LimitOverrunError,
        ):
            pass
        except Exception as e:  # Really doesn't matter if this fails, so fine with debug level.
            self.logger.debug(
                f"Something went wrong checking if `{self.qualified_name}` cog is up to date.",
                exc_info=e,
            )
        # Add SharedCog.
        if self.qualified_name != "StarBot_Utils":
            try:
                old_cog = await self.bot.remove_cog("StarBot_Utils")
                StarBot_Utils = SharedCog(self.bot)
                try:
                    if getattr(old_cog, "sentry", None) is not None:
                        StarBot_Utils.sentry = old_cog.sentry
                        StarBot_Utils.sentry.cog = StarBot_Utils
                    StarBot_Utils.loops = old_cog.loops
                except AttributeError:
                    pass
                await self.bot.add_cog(
                    StarBot_Utils, override=True
                )  # `override` shouldn't be required...
            except discord.ClientException:  # Cog already loaded.
                pass
            except Exception as e:
                self.logger.debug("Error when adding the `StarBot_Utils` cog.", exc_info=e)
            else:
                await StarBot_Utils.sentry.maybe_send_owners(self)
        # Count this cog (anonymous stats).
        StarBot_Utils = self.bot.get_cog("StarBot_Utils")
        counted_cogs = await StarBot_Utils.config.counted_cogs()
        if self.qualified_name not in counted_cogs:
            try:
                async with aiohttp.ClientSession(raise_for_status=True) as session:
                    async with session.get(
                        f"https://api.counterapi.dev/v1/StarCogs/{self.qualified_name}/up"
                    ):
                        pass
            except Exception as e:
                pass
            else:
                counted_cogs.append(self.qualified_name)
                await StarBot_Utils.config.counted_cogs.set(counted_cogs)
        # Modify hybrid commands.
        await CogsUtils.add_hybrid_commands(bot=self.bot, cog=self)

    async def cog_unload(self) -> None:
        # Close logger.
        CogsUtils.close_logger(self.logger)
        # Stop loops.
        for loop in self.loops.copy():
            if self.qualified_name == "StarBot_Utils" and loop.name == "Sentry Helper":
                continue
            await loop.execute()  # Maybe is it a loop who save data... Might execute it a last time.
            loop.stop_all()
        # Stop views.
        for view in self.views.values():
            if not view.is_finished():
                await view.on_timeout()
                view.stop()
            try:
                self.bot.persistent_views.remove(view)
            except ValueError:
                pass
        self.views.clear()
        # Remove SharedCog.
        StarBot_Utils: SharedCog = self.bot.get_cog("StarBot_Utils")
        if StarBot_Utils is not None:
            if StarBot_Utils.sentry is not None:
                await StarBot_Utils.sentry.cog_unload(self)
            if not CogsUtils.at_least_one_cog_loaded(self.bot):
                try:
                    discord.utils.get(StarBot_Utils.loops, name="Sentry Helper").stop_all()
                except ValueError:
                    pass
                await self.bot.remove_cog("StarBot_Utils")

    def format_help_for_context(self, ctx: commands.Context) -> str:
        """Thanks Simbad!"""
        text = super().format_help_for_context(ctx)
        s = "s" if len(self.__authors__) > 1 else ""
        text = bold(
            f"{text}"
            f"\n{DOT} | Author{s}: {humanize_list(self.__authors__)}"
            f"\n{DOT} | Cog version: {self.__version__}"
            f"\n{DOT} | Cog commit: `{self.__commit__}`"
            f"\n{DOT} | Repo name: {self.__repo_name__}"
            f"\n{DOT} | Utils version: {self.__utils_version__}\n"
        )
        if self.qualified_name not in ("StarBot_Utils"):
            text += bold(
                f"\n{DOT} | Cog documentation: https://docs.prismbot.icu/en/latest/cog_{self.qualified_name.lower()}.html"
                f"\n{DOT} | Translate my cogs: https://crowdin.com/project/Star-Cogs"
                f"\n{DOT} | Donate: https://ko-fi.com/starcogs"
                f"\n{DOT} | Support Server: https://discord.gg/ggQ7GcPm7P"
            )
        return text

    async def red_delete_data_for_user(self, *args, **kwargs) -> None:
        """Nothing to delete."""
        return

    async def red_get_data_for_user(self, *args, **kwargs) -> typing.Dict[str, typing.Any]:
        """Nothing to get."""
        return {}

    async def cog_before_invoke(self, ctx: commands.Context) -> Context:
        if isinstance(ctx.command, commands.Group):
            view = ctx.view
            previous = view.index
            view.skip_ws()
            trigger = view.get_word()
            invoked_subcommand = ctx.command.all_commands.get(trigger, None)
            view.index = previous
            if invoked_subcommand is not None or not ctx.command.invoke_without_command:
                return
        context: commands.Context = await Context.from_context(ctx)
        if getattr(ctx.command, "__is_dev__", False):
            await unsupported(ctx)
        if context.interaction is None:
            for index, arg in enumerate(ctx.args.copy()):
                if isinstance(arg, commands.Context):
                    ctx.args[index] = context
        else:
            if context.command.__commands_is_hybrid__ and hasattr(context.command, "app_command"):
                __do_call = getattr(context.command.app_command, "_do_call")

                async def _do_call(interaction, params):
                    await __do_call(interaction=context, params=params)

                setattr(context.command.app_command, "_do_call", _do_call)
            try:
                await context.interaction.response.defer(ephemeral=False, thinking=True)
            except (discord.InteractionResponded, discord.NotFound):
                pass
        # Typing automatically.
        if (
            ctx.cog.qualified_name not in ("CmdChannel", "Sudo")
            and (
                not (isinstance(getattr(ctx.cog, "settings", None), Settings))
                or ctx.command not in ctx.cog.settings.commands.values()
            )
            and ctx.command.qualified_name != "devutils stoptyping"
        ):
            context._typing = context.channel.typing()
            try:
                await context._typing.__aenter__()
            except discord.InteractionResponded:
                pass
        return context

    async def cog_after_invoke(self, ctx: commands.Context) -> Context:
        if isinstance(ctx.command, commands.Group) and (
            ctx.invoked_subcommand is not None or not ctx.command.invoke_without_command
        ):
            return
        context: commands.Context = await Context.from_context(ctx)
        if (
            hasattr(context, "_typing")
            and hasattr(context._typing, "task")
            and hasattr(context._typing.task, "cancel")
        ):
            context._typing.task.cancel()
        if context.command_failed:
            await context.tick(reaction="❌")
        elif not (
            getattr(ctx.cog, "qualified_name", None) == "Dev" and ctx.command.qualified_name in ("eval", "debug", "eshell")
            or getattr(ctx.cog, "qualified_name", None) == "LinkQuoter" and getattr(ctx, "__is_mocked__", False)
        ):
            await context.tick()
        # from .menus import Menu
        # await Menu(pages=str("\n".join([str((x.function, x.frame)) for x in __import__("inspect").stack(30)])), lang="py").start(context)
        return context

    async def cog_command_error(self, ctx: commands.Context, error: Exception) -> None:
        StarBot_Utils = ctx.bot.get_cog("StarBot_Utils")
        is_command_error = isinstance(
            error, (commands.CommandInvokeError, commands.HybridCommandError)
        )
        if is_command_error and isinstance(
            error.original, discord.Forbidden
        ):  # Error can be changed into `commands.BotMissingPermissions` or not.
            e = verbose_forbidden_exception(ctx, error.original)
            if e is not None and isinstance(e, commands.BotMissingPermissions):
                error = e
                is_command_error = False

        if is_command_error:
            uuid = uuid4().hex
            no_sentry = StarBot_Utils is None or getattr(StarBot_Utils, "sentry", None) is None
            if not no_sentry:
                StarBot_Utils.sentry.last_errors[uuid] = {"ctx": ctx, "error": error}
            if isinstance(ctx.command, discord.ext.commands.HybridCommand):
                _type = "[hybrid|text]" if ctx.interaction is None else "[hybrid|slash]"
            elif ctx.interaction is not None:
                _type = "[slash]"
            else:
                _type = "[text]"
            message = await ctx.bot._config.invoke_error_msg()
            if not message:
                message = f"Error in {_type} command '{ctx.command.qualified_name}'."
                if ctx.author.id in ctx.bot.owner_ids:
                    message += (
                        " Check your console or logs for details. If necessary, please inform the"
                        " creator of the cog in which this command is located. Thank you."
                    )
                message = inline(message)
            else:
                message = message.replace("{command}", ctx.command.qualified_name)
            if (
                not no_sentry
                and getattr(StarBot_Utils.sentry, "display_sentry_manual_command", True)
                and await StarBot_Utils.senderrorwithsentry.can_run(ctx)
                and not getattr(StarBot_Utils.senderrorwithsentry, "__is_dev__", False)
            ):
                message += "\n" + inline(
                    "You can send this error to the developer by running the following"
                    f" command:\n{ctx.prefix}StarBot_Utils senderrorwithsentry {uuid}"
                )
            await ctx.send(message)
            asyncio.create_task(ctx.bot._delete_delay(ctx))
            self.logger.exception(
                f"Exception in {_type} command '{ctx.command.qualified_name}'.",
                exc_info=error.original,
            )
            exception_log = f"Exception in {_type} command '{ctx.command.qualified_name}':\n"
            exception_log += "".join(
                traceback.format_exception(type(error), error, error.__traceback__)
            )
            exception_log = CogsUtils.replace_var_paths(exception_log)
            ctx.bot._last_exception = exception_log
            if not no_sentry:
                await StarBot_Utils.sentry.send_command_error(ctx, error)
        elif isinstance(error, commands.UserFeedbackCheckFailure):
            if error.message:
                message = error.message
                message = warning(message)
                await ctx.send(
                    message,
                    delete_after=3 if "delete_after" in error.args else None,
                    allowed_mentions=discord.AllowedMentions.none(),
                )
        elif isinstance(error, commands.CheckFailure) and not isinstance(
            error, commands.BotMissingPermissions
        ):
            if ctx.interaction is not None:
                await ctx.send(
                    inline("You are not allowed to execute this command in this context."),
                    ephemeral=True,
                )
        else:
            await ctx.bot.on_command_error(ctx, error=error, unhandled_by_cog=True)


def verbose_forbidden_exception(
    ctx: commands.Context, error: discord.Forbidden
) -> commands.BotMissingPermissions:  # A little useless now.
    if not isinstance(error, discord.Forbidden):
        return ValueError(error)
    method = error.response.request_info.method
    url = str(error.response.request_info.url)
    url = url[len(discord.http.Route.BASE) :]
    url = url.split("?")[0]
    url = re.sub(r"\b\d{17,20}\b", "{snowflake}", url)
    key = f"{method.upper()} {url}"
    end_points = {
        "GET /guilds/{guild.id}/audit-logs": ["VIEW_AUDIT_LOG"],
        "GET /guilds/{guild.id}/auto-moderation/rules": ["MANAGE_GUILD"],
        "GET /guilds/{guild.id}/auto-moderation/rules/{auto_moderation_rule.id}": ["MANAGE_GUILD"],
        "POST /guilds/{guild.id}/auto-moderation/rules": ["MANAGE_GUILD"],
        "PATCH /guilds/{guild.id}/auto-moderation/rules/{auto_moderation_rule.id}": [
            "MANAGE_GUILD"
        ],
        "DELETE /guilds/{guild.id}/auto-moderation/rules/{auto_moderation_rule.id}": [
            "MANAGE_GUILD"
        ],
        "PATCH /channels/{channel.id}": ["MANAGE_CHANNELS"],  # &! MANAGE_THREADS
        "DELETE /channels/{channel.id}": ["MANAGE_CHANNELS"],  # &! MANAGE_THREADS
        "GET /channels/{channel.id}/messages": [
            "VIEW_CHANNEL",
            "READ_MESSAGE_HISTORY",
        ],  # empty messages list
        "GET /channels/{channel.id}/messages/{message.id}": [
            "VIEW_CHANNEL",
            "READ_MESSAGE_HISTORY",
        ],
        "POST /channels/{channel.id}/messages": [
            "VIEW_CHANNEL",
            "SEND_MESSAGES",
        ],  # [SEND_TTS_MESSAGES (tts), READ_MESSAGE_HISTORY (reply)]
        "POST /channels/{channel.id}/messages/{message.id}/crosspost": [
            "MANAGE_MESSAGES"
        ],  # not own message
        "PUT /channels/{channel.id}/messages/{message.id}/reactions/{emoji}/@me": [
            "ADD_REACTIONS"
        ],
        "DELETE /channels/{channel.id}/messages/{message.id}/reactions/{emoji}/@me": [],
        "DELETE /channels/{channel.id}/messages/{message.id}/reactions/{emoji}/{user.id}": [
            "MANAGE_MESSAGES"
        ],
        "GET /channels/{channel.id}/messages/{message.id}/reactions/{emoji}": [],
        "DELETE /channels/{channel.id}/messages/{message.id}/reactions": ["MANAGE_MESSAGES"],
        "DELETE /channels/{channel.id}/messages/{message.id}/reactions/{emoji}": [
            "MANAGE_MESSAGES"
        ],
        "PATCH /channels/{channel.id}/messages/{message.id}": [],
        "DELETE /channels/{channel.id}/messages/{message.id}": ["MANAGE_MESSAGES"],
        "POST /channels/{channel.id}/messages/bulk-delete": [
            "VIEW_CHANNEL",
            "READ_MESSAGE_HISTORY",
            "MANAGE_MESSAGES",
        ],
        "PUT /channels/{channel.id}/permissions/{overwrite.id}": ["MANAGE_ROLES"],
        "GET /channels/{channel.id}/invites": ["MANAGE_CHANNELS"],
        "POST /channels/{channel.id}/invites": ["CREATE_INSTANT_INVITE"],
        "DELETE /channels/{channel.id}/permissions/{overwrite.id}": ["MANAGE_ROLES"],
        "POST /channels/{channel.id}/followers": ["MANAGE_WEBHOOKS"],
        "POST /channels/{channel.id}/typing": ["VIEW_CHANNEL", "SEND_MESSAGES"],
        "GET /channels/{channel.id}/pins": ["VIEW_CHANNEL", "VIEW_MESSAGE_HISTORY"],
        "PUT /channels/{channel.id}/pins/{message.id}": ["MANAGE_MESSAGES"],
        "DELETE /channels/{channel.id}/pins/{message.id}": ["MANAGE_MESSAGES"],
        "POST /channels/{channel.id}/messages/{message.id}/threads": ["CREATE_PUBLIC_THREADS"],
        "POST /channels/{channel.id}/threads": [
            "CREATE_PUBLIC_THREADS",
            "CREATE_PRIVATE_THREADS",
            "SEND_MESSAGES",
        ],
        "PUT /channels/{channel.id}/thread-members/@me": [],
        "DELETE /channels/{channel.id}/thread-members/@me": [],
        "DELETE /channels/{channel.id}/thread-members/{user.id}": ["MANAGE_THREADS"],
        "GET /channels/{channel.id}/thread-members/{user.id}": [],
        "GET /channels/{channel.id}/thread-members": [],
        "GET /channels/{channel.id}/threads/archived/public": [
            "VIEW_CHANNEl",
            "READ_MESSAGE_HISTORY",
        ],
        "GET /channels/{channel.id}/threads/archived/private": [
            "VIEW_CHANNEL",
            "READ_MESSAGE_HISTORY",
            "MANAGE_THREADS",
        ],
        "GET /channels/{channel.id}/users/@me/threads/archived/private": [
            "VIEW_CHANNEL",
            "READ_MESSAGE_HISTORY",
        ],
        "GET /guilds/{guild.id}/emojis": [],
        "GET /guilds/{guild.id}/emojis/{emoji.id}": [],
        "POST /guilds/{guild.id}/emojis": ["MANAGE_EMOJIS_AND_STICKERS"],
        "PATCH /guilds/{guild.id}/emojis/{emoji.id}": ["MANAGE_EMOJIS_AND_STICKERS"],
        "DELETE /guilds/{guild.id}/emojis/{emoji.id}": ["MANAGE_EMOJIS_AND_STICKERS"],
        "GET /guilds/{guild.id}/preview": [],
        "PATCH /guilds/{guild.id}": ["MANAGE_GUILD"],
        "DELETE /guilds/{guild.id}": [],
        "GET /guilds/{guild.id}/channels": [],
        "POST /guilds/{guild.id}/channels": ["MANAGE_CHANNELS"],
        "PATCH /guilds/{guild.id}/channels": ["MANAGE_CHANNELS"],
        "GET /guilds/{guild.id}/threads/active": [],
        "GET /guilds/{guild.id}/members/{user.id}": [],
        "GET /guilds/{guild.id}/members": [],
        "GET /guilds/{guild.id}/members/search": [],
        "PUT /guilds/{guild.id}/members/{user.id}": [],
        "PATCH /guilds/{guild.id}/members/{user.id}": ["MANAGE_MEMBERS", "MOVE_MEMBERS"],
        "PATCH /guilds/{guild.id}/members/@me": [],
        "PUT /guilds/{guild.id}/members/{user.id}/roles/{role.id}": ["MANAGE_ROLES"],
        "DELETE /guilds/{guild.id}/members/{user.id}/roles/{role.id}": ["MANAGE_ROLES"],
        "DELETE /guilds/{guild.id}/members/{user.id}": ["KICK_MEMBERS"],
        "GET /guilds/{guild.id}/bans": ["BAN_MEMBERS"],
        "GET /guilds/{guild.id}/bans/{user.id}": ["BAN_MEMBERS"],
        "PUT /guilds/{guild.id}/bans/{user.id}": ["BAN_MEMBERS"],
        "DELETE /guilds/{guild.id}/bans/{user.id}": ["BAN_MEMBERS"],
        "GET /guilds/{guild.id}/roles": [],
        "POST /guilds/{guild.id}/roles": ["MANAGE_ROLES"],
        "PATCH /guilds/{guild.id}/roles": ["MANAGE_ROLES"],
        "PATCH /guilds/{guild.id}/roles/{role.id}": ["MANAGE_ROLES"],
        "POST /guilds/{guild.id}/mfa": ["MANAGE_GUILD"],
        "DELETE /guilds/{guild.id}/roles/{role.id}": ["MANAGE_ROLES"],
        "GET /guilds/{guild.id}/prune": ["KICK_MEMBERS"],
        "POST /guilds/{guild.id}/prune": ["KICK_MEMBERS"],
        "GET /guilds/{guild.id}/regions": [],
        "GET /guilds/{guild.id}/invites": ["MANAGE_GUILD"],
        "GET /guilds/{guild.id}/integrations": ["MANAGE_GUILD"],
        "DELETE /guilds/{guild.id}/integrations/{integration.id}": ["MANAGE_GUILD"],
        "GET /guilds/{guild.id}/widget": ["MANAGE_GUILD"],
        "PATCH /guilds/{guild.id}/widget": ["MANAGE_GUILD"],
        "GET /guilds/{guild.id}/widget.json": [],
        "GET /guilds/{guild.id}/vanity-url": ["MANAGE_GUILD"],
        "GET /guilds/{guild.id}/widget.png": [],
        "GET /guilds/{guild.id}/welcome-screen": ["MANAGE_GUILD"],
        "PATCH /guilds/{guild.id}/welcome-screen": ["MANAGE_GUILD"],
        "PATCH /guilds/{guild.id}/voice-states/@me": ["MUTE_MEMBERS", "REQUEST_TO_SPEAK"],
        "PATCH /guilds/{guild.id}/voice-states/{user.id}": ["MUTE_MEMBERS"],
        "GET /guilds/{guild.id}/scheduled-events": [],
        "POST /guilds/{guild.id}/scheduled-events": ["MANAGE_EVENTS"],
        "GET /guilds/{guild.id}/scheduled-events/{guild_scheduled_event.id}": [],
        "PATCH /guilds/{guild.id}/scheduled-events/{guild_scheduled_event.id}": ["MANAGE_EVENTS"],
        "DELETE /guilds/{guild.id}/scheduled-events/{guild_scheduled_event.id}": ["MANAGE_EVENTS"],
        "GET /guilds/{guild.id}/scheduled-events/{guild_scheduled_event.id}/users": [],
        "GET /guilds/templates/{template.code}": [],
        "POST /guilds/templates/{template.code}": [],
        "GET /guilds/{guild.id}/templates": ["MANAGE_GUILD"],
        "POST /guilds/{guild.id}/templates": ["MANAGE_GUILD"],
        "PUT /guilds/{guild.id}/templates/{template.code}": ["MANAGE_GUILD"],
        "PATCH /guilds/{guild.id}/templates/{template.code}": ["MANAGE_GUILD"],
        "DELETE /guilds/{guild.id}/templates/{template.code}": ["MANAGE_GUILD"],
        "GET /invites/{invite.code}": [],
        "DELETE /invites/{invite.code}": ["MANAGE_CHANNELS"],
        "POST /stage-instances": ["MANAGE_CHANNELS"],
        "GET /stage-instances/{channel.id}": [],
        "PATCH /stage-instances/{channel.id}": ["MANAGE_CHANNELS"],
        "DELETE /stage-instances/{channel.id}": ["MANAGE_CHANNELS"],
        "GET /stickers/{sticker.id}": [],
        "GET /sticker-packs": [],
        "GET /guilds/{guild.id}/stickers": ["MANAGE_EMOJIS_AND_STICKERS"],
        "GET /guilds/{guild.id}/stickers/{sticker.id}": ["MANAGE_EMOJIS_AND_STICKERS"],
        "POST /guilds/{guild.id}/stickers": ["MANAGE_EMOJIS_AND_STICKERS"],
        "PATCH /guilds/{guild.id}/stickers/{sticker.id}": ["MANAGE_EMOJIS_AND_STICKERS"],
        "DELETE/guilds/{guild.id}/stickers/{sticker.id}": ["MANAGE_EMOJIS_AND_STICKERS"],
        "GET /users/@me": [],
        "GET /users/{user.id}": [],
        "PATCH /users/@me": [],
        "GET /users/@me/guilds": [],
        "GET /users/@me/guilds/{guild.id}/member": [],
        "DELETE /users/@me/guilds/{guild.id}": [],
        "POST /users/@me/channels": [],
        "GET /users/@me/connections": [],
        "GET /users/@me/applications/{application.id}/role-connection": [],
        "PUT /users/@me/applications/{application.id}/role-connection": [],
        "GET /voice/regions": [],
        "POST /channels/{channel.id}/webhooks": ["MANAGE_WEBHOOKS"],
        "GET /channels/{channel.id}/webhooks": ["MANAGE_WEBHOOKS"],
        "GET /guilds/{guild.id}/webhooks": ["MANAGE_WEBHOOKS"],
        "GET /webhooks/{webhook.id}": [],
        "GET /webhooks/{webhook.id}/{webhook.token}": [],
        "PATCH /webhooks/{webhook.id}": ["MANAGE_WEBHOOKS"],
        "PATCH /webhooks/{webhook.id}/{webhook.token}": ["MANAGE_WEBHOOKS"],
        "DELETE /webhooks/{webhook.id}": ["MANAGE_WEBHOOKS"],
        "DELETE /webhooks/{webhook.id}/{webhook.token}": ["MANAGE_WEBHOOKS"],
        "POST /webhooks/{webhook.id}/{webhook.token}": [],
        "POST /webhooks/{webhook.id}/{webhook.token}/slack": [],
        "POST /webhooks/{webhook.id}/{webhook.token}/github": [],
        "GET /webhooks/{webhook.id}/{webhook.token}/messages/{message.id}": [],
        "PATCH /webhooks/{webhook.id}/{webhook.token}/messages/{message.id}": [],
        "DELETE /webhooks/{webhook.id}/{webhook.token}/messages/{message.id}": [],
    }

    class FakeObject:
        id: str = "{snowflake}"
        token: str = "{snowflake}"
        code: str = "{snowflake}"

        def __str__(self):
            return "{snowflake}"

    class FakeDict(dict):
        def __getitem__(self, *args, **kwargs):
            return FakeObject()

    end_points = {_key.format_map(FakeDict()): _value for _key, _value in end_points.items()}
    if key not in end_points:
        return None
    _permissions = end_points[key]
    permissions = {}
    for permission in _permissions:
        if permission.lower() not in discord.Permissions.VALID_FLAGS:
            continue
        if getattr(ctx.bot_permissions, permission.lower()):
            continue
        permissions[permission.lower()] = True
    return (
        commands.BotMissingPermissions(discord.Permissions(**permissions)) if permissions else None
    )
